import time
from witch_msa.configs import Configs, tqdm_styles
from tqdm import tqdm
import itertools

import concurrent.futures

'''
A class defining a generic task object that will be used for job submission 
'''
class MyTask(object):
    # (required) the list of parameters used in the task
    # (optional) parent(s) of the task
    #               (i.e., the other tasks that depends on this task)
    # (optional) children of this task
    #               (i.e., the other tasks that this task depends on)
    def __init__(self, *args, **kwargs):
        self.args = tuple(*args)

        _valid = ['parent', 'children']
        for k, v in kwargs.items():
            if k in _valid:
                if isinstance(v, MyTask):
                    setattr(self, k, [v])
                elif isinstance(v, list):
                    setattr(self, k, v)
                else:
                    raise TypeError(type(v))
    
    def get_args(self):
        return self.args
    
    # currently unused
    def get_parent(self):
        if 'parent' in self.__dict__:
            return self.parent
        else:
            return None
    
    # currently unused
    def get_children(self):
        if 'children' in self.__dict__:
            return self.children
        else:
            return None

'''
Helper function to convert a list of lists of arguments to a list of MyTask objects
Assumption: all elements of args are of the same length
Return:     a generator of MyTask
'''
def getTasks(*args):
    total_length = len(args[0])
    for i in range(total_length):
        _args = [x[i] for x in args]
        yield MyTask(_args)

'''
Variant of getTasks to use given index positions to select
Also append the index to each yielded element at back
'''
def getTasksWithIndexes(indexes, *args):
    for i in indexes:
        _args = [x[i] for x in args] + [i]
        yield MyTask(_args)

'''
Helper function to handle a single future object with any return values.
Run additional callbacks with the return values and the additional callback
arguments supplemented.
Return:     the runtime to run the additional callback
'''
def handleFuture(future, success, ignored, retry, i_retry,
        callback_func, callback_args):
    s1 = time.time()
    ret = future.result()
    
    # first four fields of any callbacks will be: success=<list>,
    # ignored=<list>, retry=<list>, i_retry=<int>
    if callback_func:
        callback_func(success, ignored, retry, i_retry, 
                *ret, *callback_args)
        return time.time() - s1
    else:
        return 0.

'''
Helper function to run tasks defined by a list of MyTask objects
Required:   the function that defines the task
            the process pool to submit to
            a generator of MyTask objects
            the number of MyTask objects
Optional:   max_concurrent_jobs=<int>   # default will submit all tasks at once
            i_retry=<int>               # number of retry for failed tasks
            callback_func=<function>    # callback to run after a future is handled
                                          should take future return values as the
                                          first set of arguments
            callback_args=<*args>       # additional arguments for the callback
Return:     success, ignored, retry
            total runtime (seconds) for handling/running the tasks
'''
def runTasks(func, pool, mytasks, num_tasks, **kwargs):
    handle_runtime = 0.
    max_concurrent_jobs = kwargs.get('max_concurrent_jobs', num_tasks)
    i_retry = kwargs.get('i_retry', 0) 
    callback_func = kwargs.get('callback_func', None)
    callback_args = kwargs.get('callback_args', [])

    success, ignored, retry = [], [], []
    with tqdm(total=num_tasks, **tqdm_styles) as pbar:
        futures = {
                #pool.submit(func, *task.get_args()): task.get_id()
                pool.submit(func, *task.get_args()): task.get_parent()
                for task in itertools.islice(mytasks, max_concurrent_jobs)
        }
        while futures:
            # wait for the next future to complete
            done, _ = concurrent.futures.wait(
                    futures, return_when=concurrent.futures.FIRST_COMPLETED)
            
            for future in done:
                # depending on kwargs, allow re-adding some failed tasks back
                # to queue
                handle_runtime += handleFuture(future, success, ignored, retry,
                        i_retry, callback_func, callback_args)
                _ = futures.pop(future)
            pbar.update(len(done))

            # schedule the next batch of tasks, no more than the number of tasks
            # that just finished
            for task in itertools.islice(mytasks, len(done)):
                future = pool.submit(func, *task.get_args())
                futures[future] = task.get_parent()
    return success, ignored, retry, handle_runtime



'''
Created on Sep 29, 2020

@author: Vlad
'''

import os
import importlib
import json
import traceback
from configuration import Configs
from tasks import controller

'''
Tasks are self-contained, parallelizable units of work. 
For example, running MAFFT, compressing a subalignment, etc..
Primarily used to thread-parallelize MAFFT runs and node-parallelize subalignment operations.
Saved as JSON in task files, which are then read back by computing nodes with available threads.
This also serves the purpose of allowing aborted MAGUS runs to pick up where they left off.
'''

class Task:
    
    functionModuleMap = {"runCommand" : "tools.external_tools",
                         "runAlignmentTask" : "align.aligner",
                         "recordGapCounts" : "align.merge.alignment_writer",
                         "buildInducedSubalignment" : "align.merge.alignment_writer",
                         "compressSubalignment" : "align.merge.alignment_writer"}
    
    def __init__(self, taskType, outputFile, taskArgs, **kwargs):
        self.taskType = taskType
        self.outputFile = outputFile
        self.taskArgs = taskArgs
        
        for attr in kwargs:
            vars(self)[attr] = kwargs.get(attr)    
        self.attributes =  list(vars(self).keys())
        
        self.isFinished = False
        self.future = None
        self.json = self.toJson()   
    
    def submitTask(self):
        submitTasks([self])
    
    def awaitTask(self):
        awaitTasks([self])
        
    def submitAndAwaitTask(self):
        self.submitTask()
        self.awaitTask()
        
    def run(self):
        try:
            if not os.path.exists(self.outputFile):
                Configs.log("Running a task, output file: {}".format(self.outputFile))
                mod = importlib.import_module(Task.functionModuleMap[self.taskType])
                func = getattr(mod, self.taskType)
                func(**self.taskArgs)
                Configs.log("Completed a task, output file: {}".format(self.outputFile))
            else:
                Configs.log("File already exists: {}".format(self.outputFile))
        except Exception as exc:
            Configs.error("Task for {} threw an exception:\n{}".format(self.outputFile, exc))
            Configs.error(traceback.format_exc())
            raise
        finally:
            self.isFinished = True
        
    def checkFinished(self):
        if not self.isFinished:
            return False
        if self.future is not None:
            self.future.result()
        return True
        
    def toJson(self):
        mapper = {attr : getattr(self, attr) for attr in self.attributes}
        return json.dumps(mapper)
    
    def __eq__(self, other):
        if isinstance(other, Task):
            return self.outputFile == other.outputFile
        return NotImplemented

    def __hash__(self):
        return hash(self.outputFile)


def asCompleted(tasks):
    yield from controller.asCompleted(tasks)
            
def awaitTasks(tasks):
    controller.awaitTasks(tasks)
    
def submitTasks(tasks):
    controller.submitTasks(tasks)
    


'''
Created on 10.28.2021 by Chengze Shen

Merger of all query alignments to form the final alignment. The merging step
is exactly the same one from PASTA and UPP (by transitivity).
'''

import os, sys, re
import time
from witch_msa.configs import Configs
from witch_msa.helpers.alignment_tools import Alignment, read_fasta, \
        CompactAlignment, compact, ExtendedAlignment 
from functools import partial
#from concurrent.futures.process import ProcessPoolExecutor
from math import ceil

'''
function to merge a set of input paths (to alignments) sequentially
'''
def sequential_merger(queries, inpaths):
    init_index = 0
    while init_index < len(inpaths) and inpaths[init_index] == 'skipped':
        init_index += 1
    init_aln = Alignment(); init_aln.read_file_object(queries[init_index])
    new_aln = compact(init_aln)
    for i in range(init_index + 1, len(inpaths)):
        inpath = inpaths[i]
        
        # skip these ones
        if inpath == 'skipped':
            continue
        frag_aln = Alignment(); frag_aln.read_file_object(inpath)
        new_aln.merge_in(compact(frag_aln))
    return new_aln

'''
function to merge all subalignments to one alignment and with all singletons
in queries collapsed (in lower cases). This is the same behavior as UPP.
An additional "masked" version of the final alignment with all lower cases
removed will also be written to disk.
'''
def mergeAlignmentsCollapsed(backbone_alignment_path, queries,
        renamed_taxa, pool):
    Configs.log('(UPP-style merging) Merging all GCM subproblems ' \
            'with transitivity and singletons from queries collapsed...')
    start = time.time()
    outpath = Configs.output_path
    masked_outpath = Configs.output_path + '.masked'
    if not (len(queries) > 0):
        print('No query alignment provided to merger!')
        exit(1)

    # read in all backbone sequences/alignment
    full_aln = ExtendedAlignment([])
    full_aln.read_file_object(backbone_alignment_path)
    full_aln.from_string_to_bytearray()
    
    # read in queries so that insertions are marked
    #backbone_keys = {x: 1 for x in full_aln.keys()}
    #func = partial(getQueryAlignment, backbone_keys)
    #queries = list(pool.map(func, inpaths))

    # merge all queries to the backbone
    for query in queries:
        if query != 'skipped':
            full_aln.merge_in(query, False)
            #del query
    full_aln.from_bytearray_to_string()
    
    # rename back taxa
    name_map = {v: k for k, v in renamed_taxa.items()}
    count = 0
    for name in list(name_map.keys()):
        ori_name = name_map[name]
        if name in full_aln:
            full_aln[ori_name] = full_aln[name]
            full_aln.pop(name)
            count += 1
    if count > 0:
        Configs.log('Converted {} names back to their originals'.format(count))
    Configs.log('Finished merging all GCM subproblems, output file: {}'.format(
        outpath))
    full_aln.write(outpath, 'FASTA')
    
    # write a masked version of full alignment
    full_aln.remove_insertion_columns()
    full_aln.write(masked_outpath, 'FASTA')
    Configs.log('Masked final alignment written to: {}'.format(masked_outpath))

    end = time.time()
    Configs.runtime('Time to merge all outputs (s): {}'.format(end - start))



'''
Created on 1.19.2022 by Chengze Shen

Allow backbone alignment (by MAGUS) and backbone tree estimation (by FastTree2)
'''

import os, subprocess, time
import random
from witch_msa.gcmm import * 
from witch_msa.configs import Configs, valid_attribute
from witch_msa.helpers.alignment_tools import MutableAlignment
from argparse import Namespace

'''
Backbone alignment/tree job
'''
class BackboneJob(object):
    # default setting for a backbone alignment job
    def __init__(self, backbone_path=None, query_path=None,
            backbone_tree_path=None):
        self.alignment_method = 'magus'
        self.backbone_size = None 
        self.backbone_threshold = 0.25
        self.selection_strategy = 'median_length' 

        # executable path to the alignment method
        self.alignment_path = Configs.magus_path

        self.tree_method = 'FastTree2'
        self.tree_path = Configs.fasttreepath

        self.unaligned_backbone_path = None
        self.backbone_path = backbone_path
        self.query_path = query_path
        self.backbone_tree_path = backbone_tree_path

        # magus options
        self.magus_options = Namespace()

        self.outdir = Configs.outdir + '/tree_decomp/backbone'

    # set up from Configs.backbone namespace
    def setup(self):
        if getattr(Configs, 'Backbone', None) != None:
            for k, v in Configs.Backbone.__dict__.items():
                if v:
                    setattr(self, k, v)
        if getattr(Configs, 'MAGUS', None) != None:
            for k, v in Configs.MAGUS.__dict__.items():
                if v:
                    setattr(self.magus_options, k, v)

        # default to <outdir>/tree_decomp/backbone
        if not os.path.isdir(self.outdir):
            os.makedirs(self.outdir)

        # set backbone and queries paths to default paths
        # (and if the default path has an alignment, use it)
        if not self.unaligned_backbone_path:
            self.unaligned_backbone_path = self.outdir + '/backbone.unaln.fasta'
        if not self.backbone_path:
            self.backbone_path = self.outdir + '/backbone.aln.fasta'
        if not self.query_path:
            self.query_path = self.outdir + '/queries.fasta'
        if not self.backbone_tree_path:
            self.backbone_tree_path = self.outdir + '/backbone.tre'
        
        print('\nUsing the following settings for the backbone:')
        for k, v in self.__dict__.items():
            if valid_attribute(k, v):
                if k == 'backbone_size' and v is None:
                    print('\tBackboneJob.{}: min(1000, len(taxa))'.format(k))
                else:
                    print('\tBackboneJob.{}: {}'.format(k, v))

    # split sequences to backbone/query based on the selection strategy
    def splitSequences(self, sequences):
        assert isinstance(sequences, MutableAlignment)

        seq_lengths = sorted([len(seq) for seq in sequences.values()])
        lengths = len(seq_lengths)

        if self.backbone_size is None:
            self.backbone_size = min(1000, lengths)
        else:
            self.backbone_size = int(self.backbone_size)
        Configs.log('Backbone size set to: {}'.format(self.backbone_size))
        backbone_sequences, queries = MutableAlignment(), MutableAlignment()

        if self.selection_strategy == 'median_length':
            l2 = int(lengths / 2)
            if lengths % 2 == 1 or l2 == lengths - 1:
                median_full_length = seq_lengths[l2]
            else:
                median_full_length = (seq_lengths[l2] + seq_lengths[l2+1]) / 2.0
            min_length = int(median_full_length * (1 - self.backbone_threshold))
            max_length = int(median_full_length * (1 + self.backbone_threshold))

            Configs.log('Full length sequences set to be from '
                    + '{} to {} character long'.format(min_length, max_length))

            query_names = [name for name in sequences
                    if len(sequences[name]) > max_length or 
                    len(sequences[name]) < min_length]

            if len(query_names) > 0:
                Configs.log(
                    'Detected {} sequences not within median length'.format(
                        len(query_names)))
                queries = sequences.get_hard_sub_alignment(query_names)
                [sequences.pop(i) for i in list(queries.keys())]

            if len(sequences) < self.backbone_size:
                self.backbone_size = len(sequences)
                Configs.log('Backbone resized to: {}'.format(self.backbone_size))

            sample = sorted(random.sample(
                sorted(list(sequences.keys())), self.backbone_size))
            backbone_sequences = sequences.get_hard_sub_alignment(sample)
            [sequences.pop(i) for i in list(backbone_sequences.keys())]
        elif self.selection_strategy == 'random':
            sample = sorted(random.sample(
                sorted(list(sequences.keys())), self.backbone_size))
            backbone_sequences = sequences.get_hard_sub_alignment(sample)
            [sequences.pop(i) for i in list(backbone_sequences.keys())]
        else:
            Configs.error('Unsupported selection strategy: {}'.format(
                self.selection_strategy))
            notifyError('gcmm/backbone.py')

        # at this point, the asssumption is that backbone_sequences are
        # selected, and some sequences are put in queries
        # WRITE - backbone_sequences -> local
        # MERGE - sequences + queries -> WRITE to local
        unaligned_backbone_path = self.unaligned_backbone_path
        backbone_sequences.write(unaligned_backbone_path, 'FASTA')

        query_path = self.query_path
        sequences.set_alignment(queries)
        sequences.write(query_path, 'FASTA')

        return unaligned_backbone_path, query_path, sequences 
        
    # run alignment
    def run_alignment(self):
        start = time.time()

        if self.alignment_method == 'magus':
            self.alignment_path = Configs.magus_path
        
        # some scenarios for interrupted runs
        # (1) backbone/queries are split, and backbone is aligned
        # (2) as (1) but backbone is unaligned
        # (3) backbone/queries are not split
        if os.path.exists(self.backbone_path) \
                and os.stat(self.backbone_path).st_size > 0 \
                and os.path.exists(self.query_path):
            Configs.log('Found existing backbone alignment: {}'.format(
                self.backbone_path))
            print('\nFound existing backbone alignment: {}'.format(
                self.backbone_path))
            # (1.1) query_path is empty meaning all sequences are in backbone
            #       we assume that the user did not stop the run when writing
            #       the query sequences
            if os.stat(self.query_path).st_size == 0:
                print('\nNo query sequences to align! Exiting...')
                Configs.warning('No query sequences to align. Final alignment '
                        + 'saved to {}'.format(Configs.output_path))
                os.system('cp {} {}'.format(self.backbone_path, Configs.output_path))
                exit(0)
            else:
                return self.backbone_path, self.query_path
        elif os.path.exists(self.unaligned_backbone_path) \
                and os.path.exists(self.query_path):
            pass
        else:
            # before we split the backbone/query sequences, make sure
            # the unaligned sequences (input) exist
            if not (Configs.input_path and os.path.exists(Configs.input_path)):
                Configs.error('Input sequences file does not exist')
                raise FileNotFoundError('Input sequences file does not exist')
                #notifyError('gcmm/backbone.py - BackboneJob.run_alignment()')

            # select backbone sequences
            input_sequences = MutableAlignment()
            input_sequences.read_file_object(Configs.input_path)

            self.unaligned_backbone_path, self.query_path, queries = \
                    self.splitSequences(input_sequences)

        # scenarios (2) and (3) both need to run the alignment method,
        # default output dir to <outdir>/tree_decomp/backbone/[method]_alignment
        alignment_outdir = self.outdir + '/{}_alignment'.format(
                self.alignment_method)
        logfile_name = self.outdir \
                + '/{}_alignment_log.txt'.format(self.alignment_method)
        logfile = open(logfile_name, 'w')
        stderrdata, stdoutdata = logfile, logfile

        if self.alignment_method == 'magus':
            cmd = [self.alignment_path, '--recurse', 'false',
                    '-np', str(Configs.num_cpus),
                    '-i', self.unaligned_backbone_path,
                    '-d', alignment_outdir, '-o', self.backbone_path]
            # load in any presets for MAGUS
            for k, v in self.magus_options.__dict__.items():
                if v:
                    cmd.extend(['--{}'.format(k), str(v)])
        elif self.alignment_method == 'mafft':
            cmd = [self.alignment_path, '--quiet',
                    '--thread', str(Configs.num_cpus),
                    self.unaligned_backbone_path]
            stdoutdata = open(self.backbone_path, 'w')

        print('\nRunning {} backbone alignment...'.format(self.alignment_method))
        Configs.log('Running {} backbone alignment...'.format(
            self.alignment_method))
        Configs.debug('[{}] Command used: {}'.format(
            self.alignment_method.upper(), ' '.join(cmd)))
        p = subprocess.Popen(cmd, stdout=stdoutdata, stderr=stderrdata)
        p.wait()
        if not stdoutdata.closed:
            stdoutdata.close()
        if not stderrdata.closed:
            stderrdata.close()
        if not logfile.closed:
            logfile.close()

        # check if backbone alignment is successfully generated
        if not os.path.exists(self.backbone_path):
            Configs.error('Failed to generate {} backbone alignment, '.format(
                self.alignment_method) + 'please check log at {}'.format(
                    logfile_name))
            notifyError('gcmm/backbone.py - BackboneJob.run_alignment()')

        # need to cast all character to upper for MAFFT
        if self.alignment_method == 'mafft':
            a = MutableAlignment(); a.read_file_object(self.backbone_path)
            for taxon in a.keys():
                a[taxon] = a[taxon].upper()
            a.write(self.backbone_path, 'FASTA')

        Configs.log('Finished {} backbone alignment, backbone file: {}'.format(
            self.alignment_method, self.backbone_path) +
            ', query file: {}'.format(self.query_path))
        print('\tBackbone alignment generated at {}'.format(self.backbone_path))
        print('\tQuery sequences saved at {}'.format(self.query_path))

        # PRE-MATURE END - no queries to align. All sequences aligned in
        # the backbone
        if os.stat(self.query_path).st_size == 0:
            print('\nNo query sequences to align! Exiting...')
            Configs.warning('No query sequences to align. Final alignment '
                    + 'saved to {}'.format(Configs.output_path))
            os.system('cp {} {}'.format(self.backbone_path, Configs.output_path))
            exit(0)

        Configs.runtime('Time to align the backbone (s): {}'.format(
            time.time() - start))
        return self.backbone_path, self.query_path

    # run tree estimation
    def run_tree(self):
        start = time.time()

        # some scenarios for backbone tree estimation
        # (1) backbone_tree_path exists (finished)
        # (2) backbone_tree_path does not exist
        if os.path.exists(self.backbone_tree_path) \
                and os.stat(self.backbone_tree_path).st_size > 0:
            Configs.log('Found existing backbone tree: {}'.format(
                self.backbone_tree_path))
            print('\nFound existing backbone tree: {}'.format(
                self.backbone_tree_path))
        else:
            # before estimating the tree, making sure the backbone alignment
            # exists
            if not os.path.exists(self.backbone_path):
                Configs.error('Did not find a backbone alignment when '
                        + 'estimating the backbone tree.')
                notifyError('gcmm/backbone.py - BackboneJob.run_tree()')

            # MP-version of FastTree2
            logfile_name = self.outdir \
                    + '/{}_tree_log.txt'.format(self.tree_method)
            stderrdata = open(logfile_name, 'w')
            stdoutdata = open(self.backbone_tree_path, 'w')

            if Configs.molecule is None:
                Configs.molecule = Configs.inferDataType(self.backbone_path)

            cmd = [self.tree_path, '-gtr']
            if Configs.molecule == 'dna':
                cmd.extend(['-nt'])
            cmd.extend([self.backbone_path])
            print('\nRunning {} backbone tree estimation...'.format(
                self.tree_method))
            Configs.log('Running {} backbone tree estimation...'.format(
                self.tree_method))
            Configs.debug('[{}] Command used: {}'.format(
                self.tree_method.upper(), ' '.join(cmd)))

            os.system('export OMP_NUM_THREADS={}'.format(Configs.num_cpus))
            p = subprocess.Popen(cmd, stdout=stdoutdata, stderr=stderrdata)
            p.wait()
            if not stderrdata.closed:
                stderrdata.close()
            if not stdoutdata.closed:
                stdoutdata.close()
            
            # check if backbone tree is successfully generated
            if not os.path.exists(self.backbone_tree_path):
                Configs.error('Failed to generate {} backbone tree, '.format(
                    self.tree_method) + 'please check log at {}'.format(
                        logfile_name))
                notifyError('gcmm/backbone.py - BackboneJob.run_tree()')

            Configs.log('Finished {} backbone tree, tree file: {}'.format(
                self.tree_method, self.backbone_tree_path))
            print('\tBackbone tree estimation generated at {}'.format(
                self.backbone_tree_path))

            Configs.runtime('Time to estimate the backbone tree (s): {}'.format(
                time.time() - start))

        return self.backbone_tree_path




import os, shutil
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import ArgumentParser, Namespace
from platform import platform

def find_main_config(homepath):
    with open(homepath, 'r') as f:
        _root_dir = f.read().strip()
        main_config_path = os.path.join(_root_dir, 'main.config')
        return _root_dir, main_config_path

'''
first time run, need user to initialize the main.config
if it is not installed through github (i.e., python setup.py config)
will be needed if installed through pip/pypi
'''
def init_config_file(homepath, prioritize_user_software=True):

    # initialize a home.path that points to local user main.config
    # if it exists then pass on
    if os.path.exists(homepath):
        # detecting old home.path file based on creation time
        if os.stat(homepath).st_mtime >= os.stat(__file__).st_mtime:
            return find_main_config(homepath) 
        else:
            print('Found old home.path and regenerating...')
            os.remove(homepath)
    else:
        print('Cannot find home.path: {}'.format(homepath))

    # install to user local directory
    _root_dir = input('Create main.config file at [default: ~/.witch_msa/]: ')
    if _root_dir == '':
        _root_dir = os.path.expanduser('~/.witch_msa')
    main_config_path = os.path.join(_root_dir, 'main.config')
    print('Initializing main configuration file: {}...'.format(main_config_path))

    # write to local for installation to system
    # will read in during runs to find the main.config file
    if not os.path.isdir(_root_dir):
        os.mkdir(_root_dir)
    with open(homepath, 'w') as f:
        f.write(_root_dir)

    # create main.config file at configfile using default.config
    _config_path = os.path.join(os.path.dirname(__file__), 'default.config')
    config_defaults = []
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    assert os.path.exists('{}'.format(_config_path)), \
            "default config file {} missing! Please redownload from Github\n".format(
                    _config_path)

    if os.path.exists(main_config_path):
        print('Main configuration file {} exists...'.format(main_config_path))
        print('Overwriting existing configuration file...')

    print('\n')
    # initialize main config file using default config file
    default_config = configparser.ConfigParser()
    with open(_config_path, 'r') as f:
        default_config.read_file(f)
    for section in default_config.sections():
        cparser.add_section(section)
        for k, v in default_config[section].items():
            cparser.set(section, k, v)

    # if platform is linux then we just copy the default config file
    # as main.config
    platform_name = platform()
    tools_dir = os.path.join(os.path.dirname(__file__), 'tools')
    set_sections = ['Basic', 'MAGUS']

    # copy magus directory to tools/
    magus_dir = os.path.join(tools_dir, 'magus')
    cparser.set('Basic', 'magus_path', magus_dir + '/magus.py')

    if 'macos' not in platform_name.lower():
        print('System is {}, using default config as main.config...'.format(
            platform_name))
        # use existing binaries from MAGUS subfolder (reduce redundancy of
        # duplicated binaries)
        for _section in set_sections:
            # mafftpath
            cparser.set(_section, 'mafftpath',
                    os.path.join(magus_dir, 'tools', 'mafft', 'mafft'))
            # mclpath
            cparser.set(_section, 'mclpath',
                    os.path.join(magus_dir, 'tools', 'mcl', 'bin', 'mcl'))
            # fasttreepath
            cparser.set(_section, 'fasttreepath',
                    os.path.join(magus_dir, 'tools', 'fasttree', 'FastTreeMP'))
            # hmmer packages
            for hmmer_pkg in ['hmmsearch', 'hmmalign', 'hmmbuild']:
                cparser.set(_section, '{}path'.format(hmmer_pkg),
                        os.path.join(magus_dir, 'tools', 'hmmer', hmmer_pkg))
    else:
        if 'x86' not in platform_name:
            print('Warning: system is not using x86 architecture.',
                    'Some softwares such as FastTreeMP need to be',
                    'self-provided. See {} [Basic] '.format(_config_path),
                    'section for more information.')
        print("System is {}, reconfiguring main.config...".format(platform_name))

        # configure MAGUS to use macOS compatible executables
        binaries = os.listdir(tools_dir + '/macOS')
        for binary in binaries:
            path = os.path.join(tools_dir, 'macOS', binary)
            #path = os.path.join(_macOS_dir, binary)
            for _section in set_sections:
                if 'FastTreeMP' in path:
                    cparser.set(_section, 'fasttreepath', path)
                            #self.copy_tool_to_lib('FastTreeMP', path))
                else:
                    cparser.set(_section, '{}path'.format(binary), path)
                            #self.copy_tool_to_lib(binary, path))

    # binaries from the user's environment will be used in priority
    # if they exist
    if prioritize_user_software:
        print('Detecting existing software from the user\'s environment...')
        software = ['mafft', 'mcl', 
                'hmmsearch', 'hmmalign', 'hmmbuild', 'FastTreeMP']
        print('\tDetected:\n')
        for soft in software:
            if shutil.which(soft):
                print('\t{}: {}'.format(soft, shutil.which(soft)))
                for _section in set_sections:
                    if soft == 'FastTreeMP':
                        cparser.set(_section, 'fasttreepath',
                                shutil.which(soft))
                    elif soft == 'magus':
                        cparser.set('Basic', 'magus_path',
                                shutil.which(soft))
                    else:
                        cparser.set(_section, '{}path'.format(soft),
                                shutil.which(soft))

    with open(main_config_path, 'w') as f:
        cparser.write(f)
    print('\n(Done) main.config written to {}'.format(main_config_path))
    print('If you would like to make manual changes, please directly edit {}'.format(
        main_config_path))
    exit(0)
    #return _root_dir, main_config_path



'''
Created on Nov 1, 2020

@author: Vlad
'''

from tasks.manager import TaskManager, runTask
from configuration import Configs

'''
This is where the main thread goes to submit and await tasks.
When the main thread is in a blocking wait, it can pick up a new alignment task to run.
'''

def submitTasks(tasks):
    with TaskManager.managerLock:
        checkTaskManager()
        TaskManager.submittedTasks.update(tasks)
        TaskManager.managerSignal.set()

def asCompleted(tasks):
    unfinished = tasks
    while True:
        finished, unfinished = checkWhatFinished(unfinished)
        yield from finished
        if len(unfinished) == 0:
            return
        if len(finished) == 0:
            observeTaskManager()
        
def awaitTasks(tasks):
    finished, unfinished = checkWhatFinished(tasks)
    while len(unfinished) > 0:
        observeTaskManager()
        finished, unfinished = checkWhatFinished(unfinished)
        
def checkWhatFinished(tasks):
    finished, unfinished = [], []
    for t in tasks:
        finished.append(t) if t.checkFinished() else unfinished.append(t)
    return finished, unfinished

def observeTaskManager():
    TaskManager.observerWaiting = True
    TaskManager.managerSignal.set()
    TaskManager.observerSignal.wait(10)
    with TaskManager.managerLock:
        TaskManager.observerWaiting = False
        TaskManager.observerSignal.clear()
        checkTaskManager()
        task, TaskManager.observerTask = TaskManager.observerTask, None
    if task is not None:
        runTask(task)
            
        #manager.setTaskFinished(task)

def checkTaskManager():
    if not TaskManager.managerFuture.running():
        Configs.error("Task manager is dead for some reason..")
        TaskManager.managerFuture.result()
        raise Exception("Task manager is dead for some reason..")
        



'''
Created on Aug 23, 2020

@author: Vlad
'''


from configuration import Configs

def atomizedClustering(graph):
    Configs.log("Building a fully atomized clustering..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    graph.clusters = atomizedCluster(lowerBound, upperBound)

def naiveClustering(graph):
    Configs.log("Building a naive left-justified clustering..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    graph.clusters = naiveCluster(lowerBound, upperBound)

def atomizedCluster(lowerBound, upperBound):
    clusters = []
    for j in range(len(lowerBound)):
        for i in range(lowerBound[j], upperBound[j]):
            clusters.append([i])
    return clusters

def naiveCluster(lowerBound, upperBound):
    clusters = []
    i = 0
    while True:
        cluster = []
        for j in range(len(lowerBound)):
            if lowerBound[j] + i < upperBound[j]:
                cluster.append(lowerBound[j] + i)
        if len(cluster) == 0:
            break   
        clusters.append(cluster)     
        i = i+1
    return clusters


'''
Created on May 29, 2020

@author: Vlad
'''

import os
import shutil
import time
import random

from align.decompose import decomposer
from helpers import sequenceutils, treeutils, hmmutils
from tasks import task
from tools import external_tools
from configuration import Configs

'''
Not really used for anything, may be removed in the future.
'''

def buildSubsetsKMH(context, subsetsDir):
    tempDir = os.path.join(subsetsDir, "initial_tree")
    
    Configs.log("Building KMH decomposition on {} with skeleton size {}/{}..".format(context.sequencesPath, Configs.decompositionSkeletonSize, 1000))
    time1 = time.time()
    
    initialTreePath, initialAlignPath, unusedTaxa = buildInitialTreeAlign(tempDir, context.sequencesPath) 
    
    if len(unusedTaxa) == 0:
        subsetPaths = treeutils.decomposeGuideTree(tempDir, initialAlignPath, initialTreePath, Configs.decompositionMaxSubsetSize, Configs.decompositionMaxNumSubsets)
    else:
        subsetSeedDir = os.path.join(subsetsDir, "seed_subsets")
        if not os.path.exists(subsetSeedDir):
            os.makedirs(subsetSeedDir)
        subsetSeedPaths = treeutils.decomposeGuideTree(subsetSeedDir, initialAlignPath, initialTreePath, None, Configs.decompositionMaxNumSubsets)
        subsetPaths = reassignTaxons(subsetsDir, subsetSeedPaths, context.unalignedSequences, unusedTaxa)
    
    time2 = time.time()
    Configs.log("Built KMH decomposition on {} in {} sec..".format(context.sequencesPath, time2-time1))

    return subsetPaths

def buildInitialTreeAlign(tempDir, sequencesPath):
    outputTreePath = os.path.join(tempDir, "initial_tree.tre")
    outputAlignPath = os.path.join(tempDir, "initial_align.txt")
    
    if os.path.exists(outputTreePath) and os.path.exists(outputAlignPath):
        return outputTreePath, outputAlignPath
    if os.path.exists(tempDir):
        shutil.rmtree(tempDir)
    os.makedirs(tempDir)
    
    initialAlign, unusedTaxa = decomposer.initial_tree.buildInitialAlignment(sequencesPath, tempDir, Configs.decompositionSkeletonSize, 1000)
    sequenceutils.writeFasta(initialAlign, outputAlignPath)    
    #external_tools.runRaxmlNg(outputAlignPath, tempDir, outputTreePath, 8).run()
    external_tools.runFastTree(outputAlignPath, tempDir, outputTreePath).run()

    return outputTreePath, outputAlignPath, unusedTaxa

def reassignTaxons(subsetsDir, subsetSeedPaths, sequences, unusedTaxa):
    unusedPath = os.path.join(subsetsDir, "unassigned_sequences.txt")
    sequenceutils.writeFasta(sequences, unusedPath, unusedTaxa)
    
    hmmMap = {}
    for subsetPath in subsetSeedPaths:
        hmmDir = os.path.join(os.path.dirname(subsetPath), "hmm_{}".format(os.path.basename(subsetPath)).replace(".", "_"))
        if not os.path.exists(hmmDir):
            os.makedirs(hmmDir)
        hmmMap[subsetPath] = os.path.join(hmmDir, "hmm_model.txt") 
    hmmTasks = hmmutils.buildHmms(hmmMap)
    task.submitTasks(hmmTasks)
    task.awaitTasks(hmmTasks)
    hmmPaths = [t.outputFile for t in hmmTasks]
    
    scoreFileHmmFileMap = {}
    scoreTasks = hmmutils.buildHmmScores(hmmPaths, unusedPath, scoreFileHmmFileMap)
    task.submitTasks(scoreTasks) 
    
    bestScores = {}
    taxonHmmMap = {}
    for scoreTask in task.asCompleted(scoreTasks):
        subsetScores = hmmutils.readSearchFile(scoreTask.outputFile)
        for taxon, scores in subsetScores.items():
            if scores[1] > bestScores.get(taxon, -float("inf")):
                bestScores[taxon] = scores[1]
                taxonHmmMap[taxon] = scoreFileHmmFileMap[scoreTask.outputFile]
    
    subsetTaxons = {file : [] for file in hmmPaths}
    for taxon, hmmPath in taxonHmmMap.items():
        subsetTaxons[hmmPath].append(taxon)
    for subsetPath, hmmPath in hmmMap.items():
        subset = sequenceutils.readFromFasta(subsetPath)
        for taxon in subset:
            subsetTaxons[hmmPath].append(taxon)
    
    subsetPaths = []    
    i = 1
    for hmmPath, subset in subsetTaxons.items():
        subsetPath = os.path.join(subsetsDir, "subset_{}.txt".format(i))
        subsetPaths.append(subsetPath)
        sequenceutils.writeFasta(sequences, subsetPath, subset)
        i = i + 1

    return subsetPaths


    



'''
Created on May 14, 2020

@author: Vlad
'''

import os
import time


from align.merge.graph_build.graph_builder import buildGraph
from align.merge.graph_cluster.clusterer import clusterGraph
from align.merge.graph_trace.tracer import findTrace
from align.merge.optimizer import optimizeTrace
from align.merge.alignment_writer import writeAlignment
from configuration import Configs


def mergeSubalignments(context):
    Configs.log("Merging {} subaligments..".format(len(context.subalignmentPaths)))
    time1 = time.time()  
    
    buildGraph(context)
    clusterGraph(context.graph)
    findTrace(context.graph)
    optimizeTrace(context.graph)    
    writeAlignment(context)
    
    time2 = time.time()  
    Configs.log("Merged {} subalignments into {} in {} sec..".format(len(context.subalignmentPaths), context.outputFile, time2-time1))



'''
Created on May 29, 2020

@author: Vlad
'''

import os
import shutil

from align.alignment_context import AlignmentContext
from align.decompose.decomposer import decomposeSequences 
from align.merge.merger import mergeSubalignments
from tools import external_tools
from configuration import Configs
from helpers import sequenceutils
from tasks import task

'''
Alignments are treated as "tasks", units of work that are written out to task files and 
processed as threads and/or compute nodes become available. 
MAGUS tasks will recursively generate MAGUS tasks over large subsets and MAFFT tasks over smaller subsets.
'''

def mainAlignmentTask():    
    '''
        6.1.2021 - modified by Chengze Shen
        added a new argument for backboneWeightsPath
    '''
    args = {"workingDir" : Configs.workingDir, "outputFile" : Configs.outputPath,
            "subalignmentPaths" : Configs.subalignmentPaths, "sequencesPath" : Configs.sequencesPath,
            "backbonePaths" : Configs.backbonePaths, "guideTree" : Configs.guideTree,
            "backboneWeightsPath": Configs.backboneWeightsPath}
    task = createAlignmentTask(args)
    task.submitTask()
    task.awaitTask()
    
def createAlignmentTask(args):
    return task.Task(taskType = "runAlignmentTask", outputFile = args["outputFile"], taskArgs = args)

def runAlignmentTask(**kwargs):
    '''
    The standard MAGUS task: 
    decompose the data into subsets, align each subset, and merge the subalignments.
    '''
    
    with AlignmentContext(**kwargs) as context:
        if context.sequencesPath is not None:
            Configs.log("Aligning sequences {}".format(context.sequencesPath))
        
        decomposeSequences(context)
        if Configs.onlyGuideTree:
            Configs.log("Outputting only the guide tree, as requested..")
            shutil.copyfile(os.path.join(context.workingDir, "decomposition", "initial_tree", "initial_tree.tre"), context.outputFile)
            return
        
        '''
            6.1.2021 - added by Chengze Shen
            a quick log for sanity check that the backbone weights path is correct
        '''
        if context.backboneWeightsPath:
            Configs.log("Backbone weights path is {}..".format(
                context.backboneWeightsPath))
        else:
            Configs.log("Heads up! All backbone alignments are treated equally..")

        alignSubsets(context)
        mergeSubalignments(context)

def alignSubsets(context):
    if len(context.subalignmentPaths) > 0:
        Configs.log("Subalignment paths already provided, skipping subalignments..")
        return
    
    Configs.log("Building {} subalignments..".format(len(context.subsetPaths)))
    subalignDir = os.path.join(context.workingDir, "subalignments")
    if not os.path.exists(subalignDir):
        os.makedirs(subalignDir)
        
    mafftThreshold = max(Configs.mafftSize, Configs.decompositionMaxSubsetSize, Configs.recurseThreshold)
    
    for file in context.subsetPaths:
        subset = sequenceutils.readFromFasta(file)
        subalignmentPath = os.path.join(subalignDir, "subalignment_{}".format(os.path.basename(file)))
        context.subalignmentPaths.append(subalignmentPath)
        
        if os.path.exists(subalignmentPath):
            Configs.log("Existing subalignment file detected: {}".format(subalignmentPath))       
             
        elif len(subset) <= mafftThreshold or not Configs.recurse:
            Configs.log("Subset has {}/{} sequences, aligning with MAFFT..".format(len(subset), mafftThreshold))            
            subalignmentTask = external_tools.buildMafftAlignment(file, subalignmentPath)
            context.subalignmentTasks.append(subalignmentTask)
            
        else:
            Configs.log("Subset has {}/{} sequences, recursively subaligning with MAGUS..".format(len(subset), mafftThreshold))
            subalignmentDir = os.path.join(subalignDir, os.path.splitext(os.path.basename(subalignmentPath))[0])
            subalignmentTask = createAlignmentTask({"outputFile" : subalignmentPath, "workingDir" : subalignmentDir, 
                                                    "sequencesPath" : file, "guideTree" : Configs.recurseGuideTree})   
            context.subalignmentTasks.append(subalignmentTask)
                
    task.submitTasks(context.subalignmentTasks)
    Configs.log("Prepared {} subset alignment tasks..".format(len(context.subalignmentTasks)))




'''
Created on Apr 14, 2020

@author: Vlad
'''

import os

from helpers import sequenceutils
from configuration import Configs
import threading


'''
Data structure for dealing with alignment graphs.
Subalignment columns are mapped to graph nodes, represented by integers.
Integer nodes can be converted back to corresponding subalignment columns.
Reads/writes graph and cluster files.
'''

class AlignmentGraph:
    
    def __init__(self, context):
        self.context = context
        self.workingDir = os.path.join(self.context.workingDir, "graph")
        self.graphPath = os.path.join(self.workingDir, "graph.txt")
        self.clusterPath = os.path.join(self.workingDir, "clusters.txt")
        self.tracePath = os.path.join(self.workingDir, "trace.txt")
        if not os.path.exists(self.workingDir):
            os.makedirs(self.workingDir)
        
        self.subalignmentLengths = []
        self.subsetMatrixIdx = []
        self.matSubPosMap = []
        
        self.matrixSize = 0
        self.matrix = None
        self.matrixLock = threading.Lock()
        self.nodeEdges = None
        
        self.clusters = []
        self.insertions = set()
        
    def initializeMatrix(self):
        if Configs.constrain:
            self.subalignmentLengths = [sequenceutils.readSequenceLengthFromFasta(file) for file in self.context.subalignmentPaths]
        else:
            self.subalignmentLengths = [len(self.context.unalignedSequences[s[0]].seq) for s in self.context.subalignments]
        
        self.matrixSize = sum(self.subalignmentLengths)    
        self.subsetMatrixIdx = [0] * len(self.subalignmentLengths)
        for k in range(1, len(self.subalignmentLengths)):        
            self.subsetMatrixIdx[k] = self.subsetMatrixIdx[k-1] + self.subalignmentLengths[k-1]
        
        self.matSubPosMap = [0] * self.matrixSize
        i = 0
        for k in range(len(self.subalignmentLengths)):
            for j in range(self.subalignmentLengths[k]):
                self.matSubPosMap[i] = (k, j)
                i = i + 1
        
        self.matrix = [{} for i in range(self.matrixSize)]
    
    def writeGraphToFile(self, filePath):
        with open(filePath, 'w') as textFile:
            for i in range(len(self.matrix)):
                for k in self.matrix[i]:
                    textFile.write("{} {} {}\n".format(i, k, self.matrix[i][k]))
        Configs.log("Wrote matrix to {}".format(filePath))

    def readGraphFromFile(self, filePath):
        self.matrix = [{} for i in range(self.matrixSize)]
        with open(filePath) as f:
            for line in f:
                tokens = [int(token) for token in line.strip().split()]
                self.matrix[tokens[0]][tokens[1]] = tokens[2]
        Configs.log("Read matrix from {}".format(filePath))
    
    def writeClustersToFile(self, filePath):
        with open(filePath, 'w') as textFile:
            for cluster in self.clusters:
                textFile.write("{}\n".format(" ".join([str(c) for c in cluster])))
        
    def readClustersFromFile(self, filePath):
        self.clusters = []
        with open(filePath) as f:
            for line in f:
                tokens = [int(token) for token in line.strip().split()]
                if len(tokens) > 1:
                    self.clusters.append(tokens) 
        print("Found {} clusters..".format(len(self.clusters)))
    
    def buildNodeEdgeDataStructure(self):
        Configs.log("Preparing node edge data structure..")
        k = len(self.subalignmentLengths)
        self.nodeEdges = {}
        
        for a in range(self.matrixSize):
            asub, apos = self.matSubPosMap[a]
            self.nodeEdges[a] = [[] for i in range(k)]
            for b, value in self.matrix[a].items():
                bsub, bpos = self.matSubPosMap[b] 
                if asub == bsub:
                    continue
                self.nodeEdges[a][bsub].append((b, value))
            for i in range(k):
                self.nodeEdges[a][i].sort(key = lambda pair: pair[0])
        Configs.log("Prepared node edge data structure..")
    
    def buildNodeEdgeDataStructureFromClusters(self):
        Configs.log("Preparing node edge data structure..")
        k = len(self.subalignmentLengths)
        self.nodeEdges = {}
        
        Configs.log("Using {} pre-existing clusters to simplify alignment graph..".format(len(self.clusters)))
        for a in range(self.matrixSize):
            self.nodeEdges[a] = [[] for i in range(k)]
        
        for cluster in self.clusters:
            for a in cluster:
                asub, apos = self.matSubPosMap[a]
                for b in cluster:
                    bsub, bpos = self.matSubPosMap[b]
                    if asub == bsub or b not in self.matrix[a]:
                        continue
                    value = self.matrix[a][b]
                    self.nodeEdges[a][bsub].append((b, value))
                for i in range(k):
                    self.nodeEdges[a][i].sort(key = lambda pair: pair[0])
        Configs.log("Prepared node edge data structure..")

    def cutString(self, cut):
        stringCut = list(cut)
        for i, value in enumerate(stringCut):
            stringCut[i] = value - self.subsetMatrixIdx[i]
        return stringCut

    def computeClusteringCost(self, clusters):
        cutCost = 0
        nodeClusters = {}
        
        for n, cluster in enumerate(clusters):
            for a in cluster:
                nodeClusters[a] = n
                
        clusterCounter = len(clusters)
        for a in range(self.matrixSize):
            if a not in nodeClusters:
                nodeClusters[a] = clusterCounter
                clusterCounter = clusterCounter + 1
                 
        for a in range(self.matrixSize):
            asub, apos = self.matSubPosMap[a]
            for b, value in self.matrix[a].items():
                bsub, bpos = self.matSubPosMap[b] 
                if asub != bsub and nodeClusters[a] != nodeClusters[b]:
                    cutCost = cutCost + value
    
        return int(cutCost/2) 
    
    def addSingletonClusters(self):
        newClusters = []
        
        lastIdx = list(self.subsetMatrixIdx) 
        for cluster in self.clusters:
            for a in cluster:
                #print(a)
                asub, apos = self.matSubPosMap[a]                
                for node in range(lastIdx[asub], a):
                    newClusters.append([node])
                lastIdx[asub] = a+1
            newClusters.append(cluster)
        for i in range(len(lastIdx)):
            for node in range(lastIdx[i], self.subsetMatrixIdx[i] + self.subalignmentLengths[i]):
                newClusters.append([node])
        self.clusters = newClusters
        return newClusters
    
    
    
    


'''
Created on 1.19.2022 by Chengze Shen

Main pipeline of WITCH
'''

import os, sys, math, psutil, shutil 
from witch_msa.configs import * 
from witch_msa.gcmm.algorithm import DecompositionAlgorithm, SearchAlgorithm
from witch_msa.gcmm.loader import loadSubQueries, rankBitscores, writeTempBackbone
from witch_msa.gcmm.weighting import writeWeights, writeBitscores, \
        readWeightsFromLocal, writeWeightsToLocal
from witch_msa.gcmm.aligner import alignSubQueries, alignSubQueriesNew
from witch_msa.gcmm.backbone import BackboneJob
from witch_msa.gcmm.merger import mergeAlignmentsCollapsed
from witch_msa.gcmm.results_handler import submitAndCollectFutures

# experimental, customized ProcessPoolExecutor
from witch_msa.gcmm import *

from witch_msa.helpers.alignment_tools import Alignment
from witch_msa.helpers.general_tools import memoryUsage

import multiprocessing as mp
import concurrent.futures
from multiprocessing import Lock, Queue, Manager#, Pool
from concurrent.futures.process import ProcessPoolExecutor
from functools import partial
from tqdm import tqdm

# max system recursion limit hard encoding to a large number
# a temp fix for dendropy tree recursion issues
sys.setrecursionlimit(10000)

'''
Delete all unnecessary intermediate files
'''
def clearTempFiles():
    # make an empty dir for rsync removal
    blank_dir = os.path.join(Configs.outdir, 'blank')
    if not os.path.isdir(blank_dir):
        os.makedirs(blank_dir)
    #if not Configs.keepsubalignment:
    if os.path.isdir('{}/temp'.format(Configs.outdir)):
        #os.system('rsync -a --delete {}/ {}/temp/'.format(blank_dir,
        #    Configs.outdir))
        os.system('rm -rf {}/temp'.format(Configs.outdir))

    dirs_to_remove = ['tree_decomp',
            'backbone_alignments', 'constraints',
            'search_results', 'data', 'weights', 'bitscores']
    if Configs.keep_decomposition:
        dirs_to_remove = dirs_to_remove[1:]

    for _d in dirs_to_remove:
        if os.path.isdir('{}/{}'.format(Configs.outdir, _d)):
            os.system('rsync -a --delete {}/ {}/{}/'.format(blank_dir,
                Configs.outdir, _d))
            os.system('rmdir {}/{}'.format(Configs.outdir, _d))

    if (not Configs.keepgcmtemp) \
            and os.path.isdir('{}/magus_outputs'.format(Configs.outdir)):
        os.system('rsync -a --delete {}/ {}/magus_outputs/'.format(blank_dir,
            Configs.outdir))
        os.system('rmdir {}/magus_outputs'.format(Configs.outdir))

    if os.path.isdir(blank_dir):
        os.system('rmdir {}'.format(blank_dir))

'''
Init function for a queue and get configurations for each worker
'''
def initiate_pool(args):
    buildConfigs(args)

'''
6.21.2023 - Additionally, init pool for query alignment with two dicts
          - used later
'''
def initiate_pool_query_alignment(q, args,
        subset_to_retained_columns,
        subset_to_nongaps_per_column):
    buildConfigs(args)
    alignSubQueries.q = q
    alignSubQueriesNew.subset_to_retained_columns = subset_to_retained_columns
    alignSubQueriesNew.subset_to_nongaps_per_column = \
            subset_to_nongaps_per_column

'''
Main process for WITCH 
'''
def mainAlignmentProcess(args):
    m = Manager()
    lock = m.Lock()
    #l = Lock()
    q = Queue()

    # initialize the main pool at the start so that it's not memory
    # intensive
    Configs.warning('Initializing ProcessorPoolExecutor instance...')
    pool = ProcessPoolExecutor(Configs.num_cpus,
            initializer=initiate_pool, initargs=(args,))
            #mp_context=mp.get_context('spawn'),

    # if not user provided, default to <outdir>/tree_decomp/root
    #if not Configs.hmmdir:
    #    Configs.hmmdir = Configs.outdir + '/tree_decomp/root'

    # 0) obtain the backbone alignment/tree and eHMMs
    # If no UPP eHMM directory detected, decompose from the backbone
    if not Configs.hmmdir:
        # default to <outdir>/tree_decomp/root
        Configs.hmmdir = Configs.outdir + '/tree_decomp/root'
    else:
        # if user provides a directory but it does not exist, exit program
        assert os.path.isdir(Configs.hmmdir), \
            'Provided HMM directory does not exist'

    if not os.path.isdir(Configs.hmmdir):
        # if both backbone alignment/tree are provided by the user
        if Configs.backbone_path and Configs.backbone_tree_path:
            pass
        else:
            # if missing backbone alignment, first perform an UPP-like
            # sequence split into backbone/query sets (i.e., randomly select
            # up to 1,000 sequences from 25% of the median length to be
            # backbone sequences)
            # Then, align the backbone sequences using MAGUS/PASTA
            # Finally, generate a FastTree2 backbone tree

            # jobs will only run if the corresponding paths are missing
            print('\nPerforming backbone alignment and/or tree estimation...')
            bb_job = BackboneJob(Configs.backbone_path, Configs.query_path,
                    Configs.backbone_tree_path)
            bb_job.setup()

            Configs.backbone_path, Configs.query_path = \
                    bb_job.run_alignment()
            Configs.backbone_tree_path = bb_job.run_tree()
    
        # after obtaining backbone alignment/tree, perform decomposition
        # and HMMSearches
        print('\nDecomposing the backbone tree...')
        decomp = DecompositionAlgorithm(
                Configs.backbone_path, Configs.backbone_tree_path,
                Configs.alignment_size)
        hmmbuild_paths, subset_to_retained_columns, subset_to_nongaps_per_column = \
                decomp.decomposition(lock, pool)
        print('\nPerforming all-against-all HMMSearches ' \
                'between the backbone and queries...')
        search = SearchAlgorithm(hmmbuild_paths)
        hmmsearch_paths = search.search(lock, pool)
        del decomp; del search
    else:
        # go over the given hmm directory and obtain all subset alignment
        # get their retained columns with respect to the backbone alignment
        print('\nFound existing HMM directory: {}'.format(Configs.hmmdir))        
        _dummy_search = SearchAlgorithm(None)
        backbone_path, subset_to_retained_columns, subset_to_nongaps_per_column \
                = _dummy_search.readHMMDirectory(lock, pool)
        if not Configs.backbone_path:
            Configs.backbone_path = backbone_path

    # sanity check before moving on
    # minimum required at this point:
    #   1. hmmdir   2. backbone_path    3. query_path
    assert Configs.backbone_path != None \
            and os.path.exists(Configs.backbone_path), 'backbone alignment missing'
    #assert Configs.backbone_tree_path != None \
    #        and os.path.exists(Configs.backbone_tree_path), 'backbone tree missing'
    assert Configs.query_path != None \
            and os.path.exists(Configs.query_path), 'query sequences missing'
    assert Configs.hmmdir != None \
            and os.path.isdir(Configs.hmmdir), 'eHMM directory missing'

    ####### Addition - 6.13.2023 #######
    # Create lists that record retained column indexes and the
    # number of nongap characters per column
    # of each subset alignment for in-memory merging of HMM-query alignments,
    # instead of calling GCM (the WITCH-ng's way)
    # will be passed to the initiation of process pool as initial arguments

    # close pool and re-initiate pool to run the actual query alignment part
    Configs.warning('Closing ProcessPoolExecutor instance (for backbone)...')
    pool.shutdown()
    #pool = ProcessPoolExecutor(Configs.num_cpus,
    pool = WITCHProcessPoolExecutor(Configs.num_cpus,
            initializer=initiate_pool_query_alignment,
            initargs=(q, args, subset_to_retained_columns,
                subset_to_nongaps_per_column))
            #mp_context=mp.get_context('spawn'),
    Configs.warning('ProcessPoolExecutor instance re-opened (for query alignment).')

    # 0) create a temporary (working) backbone alignment,
    # enforcing all letters to be upper-cases
    tmp_backbone_path, backbone_length = writeTempBackbone(
            Configs.outdir + '/tree_decomp/backbone', Configs.backbone_path)

    # 1) get all sub-queries
    num_seq, index_to_hmm, sid_to_query_names, \
            sid_to_query_seqs, renamed_taxa = loadSubQueries(lock, pool)

    # 2) calculate weights, if needed
    # read in existing weights.txt file if it exists
    weight_path = Configs.outdir + '/weights.txt'
    if os.path.exists(weight_path): 
        print('\nFound existing weights: {}'.format(weight_path))
        taxon_to_weights = readWeightsFromLocal(weight_path)
    else:
        ranked_bitscores = rankBitscores(index_to_hmm, renamed_taxa, lock, pool)
        if Configs.use_weight:
            print('\nCalculating weights (adjusted bit-scores)...')
            taxon_to_weights = writeWeights(index_to_hmm, ranked_bitscores, pool)
        else:
            print('\nLoading bit-scores...')
            taxon_to_weights = writeBitscores(ranked_bitscores, pool)
        del ranked_bitscores

        # 2a) if saving weights to local
        if Configs.save_weight:
            print('\n(user option) Writing weights to local...')
            writeWeightsToLocal(taxon_to_weights, Configs.outdir + '/weights.txt')

    # 3) solve each subset
    sub_alignment_paths = []
    if not os.path.isdir(Configs.outdir + '/temp'):
        os.makedirs(Configs.outdir + '/temp')

    # ProcessPoolExecutor version
    print('\nPerforming GCM alignments on query subsets...')

    queries, ignored_indexes = submitAndCollectFutures(pool, lock,
            sid_to_query_names, sid_to_query_seqs,
            num_seq, index_to_hmm, taxon_to_weights,
            tmp_backbone_path, backbone_length)

    # 4) merge all results 
    print('\nAll GCM subproblems finished! Doing merging with transitivity...')
    mergeAlignmentsCollapsed(tmp_backbone_path, queries, renamed_taxa, pool)

    # if there are any ignored indexes (queries), write them to local
    ignored_path = Configs.outdir + '/ignored_queries.fasta'
    if len(ignored_indexes) > 0:
        Configs.log('Writing {} ignored sequences to local at {}'.format(
            len(ignored_indexes), ignored_path)) 
        newname_to_oldname = {v: k for k, v in renamed_taxa.items()}
        with open(ignored_path, 'w') as f:
            for _i in ignored_indexes:
                name, seq = sid_to_query_names[_i], sid_to_query_seqs[_i]
                if name in newname_to_oldname:
                    name = newname_to_oldname[name]
                f.write('>{}\n{}\n'.format(name, seq))

    Configs.warning('Closing ProcessPoolExecutor instance...')
    pool.shutdown()
    Configs.warning('ProcessPoolExecutor instance closed.')
    if not Configs.keeptemp:
        clearTempFiles()



'''
Created on Apr 18, 2020

@author: Vlad
'''

import heapq

from configuration import Configs

'''
Resolve clusters into a trace by breaking conflicting clusters apart.
We use A* to search for the path of cluster breaks with the smallest number of clusters broken.
'''

#todo refactoring
def minClustersSearch(graph):
    Configs.log("Finding graph trace with minimum clusters heuristic search..")
    
    subsetClusters = {}
    clusterPositions = {}
    queueIdxs = {}
    clusterBreaks = {}
    maxFrontier = {}
    visitedStates = set()
    maximalCut = {}
    stateCounter = 0
    aggression = 1.0
    lastFrontierState = None
    greedy = False
    totalPairs = 0

    
    for a,cluster in enumerate(graph.clusters):
        for b in cluster:
            bsub, bpos = graph.matSubPosMap[b] 
            subsetClusters[bsub] = subsetClusters.get(bsub, []) + [(a, bpos)]  
            clusterPositions[a] = {}
            totalPairs = totalPairs + len(cluster)*(len(cluster)-1)/2
                
    
    for asub in subsetClusters:
        subsetClusters[asub].sort(key = lambda c: c[1])
        queueIdxs[asub] = 0

        maxFrontier[asub] = -1
        maximalCut[asub] = -1
        
        for i in range(len(subsetClusters[asub])):            
            a = subsetClusters[asub][i][0]
            clusterPositions[a][asub] = i
    
    heap = []
    
    startState = (0, 0, len(graph.clusters), totalPairs, stateCounter, queueIdxs, clusterBreaks, maximalCut, [], True)
    startState = developState(startState, graph, aggression, greedy, 0, subsetClusters, clusterPositions)
    heapq.heappush(heap, startState)
    
    while len(heap) > 0:
        heapCleared = False
        if len(heap) > Configs.searchHeapLimit:
            Configs.log("Heap limit {} reached.. Truncating heap to last frontier".format(Configs.searchHeapLimit)) 
            if aggression == 1:
                aggression = 1.2
                Configs.log("Increasing aggression to {}..".format(aggression))
            elif aggression < 8:
                aggression = int(aggression) * 2
                Configs.log("Increasing aggression to {}..".format(aggression))
            else:
                Configs.log("Setting search strategy to fully greedy..")
                greedy = True
                aggression = 1    
                
            heap.clear()
            visitedStates = set()
            lastFrontierState = developState(lastFrontierState, graph, aggression, greedy, 0, subsetClusters, clusterPositions)
            heapq.heappush(heap, lastFrontierState)
            heapCleared = True

        state = heapq.heappop(heap)
        heuristic, numOrdered, numLeft, pairsLeft, counter, queueIdxs, clusterBreaks, maximalCut, newClusterBreaks, safeFrontier = state
   
        if len(newClusterBreaks) == 0:
            break
        else:
            stateKey = tuple([queueIdxs[sub] for sub in subsetClusters])
            if stateKey in visitedStates:
                continue
            else:
                visitedStates.add(stateKey)
            

            newSubFrontier = True
            for asub in queueIdxs:
                if queueIdxs[asub] <= maxFrontier[asub]:
                    newSubFrontier = False
                    break
                
            if newSubFrontier:
                maxFrontier = queueIdxs
                Configs.log("Reached new search frontier")
                Configs.log(maxFrontier)
                lastFrontierState = state
                greedy = False
            
            if safeFrontier and not heapCleared:
                Configs.log("Safe frontier reached.. dumping {} from heap and resetting aggression..".format(len(heap)))
                lastFrontierState = state
                heap.clear()
                visitedStates = set()
                aggression = 1.0
                greedy = False
                
            
            nextStates = []
            for a, goodSide, badSide, crossedClusters in newClusterBreaks:
                
                g,b = len(goodSide), len(badSide)
                pairsDiff = g*(g-1)/2 + b*(b-1)/2 - (g+b)*(g+b-1)/2
                
                stateCounter = stateCounter + 1
                queueIdxsCopy = dict(queueIdxs)
                clusterBreaksCopy = dict(clusterBreaks)
                maximalCutCopy = dict(maximalCut)
                for b in goodSide:
                    bsub, bpos = graph.matSubPosMap[b]
                    clusterBreaksCopy[a, bsub] = goodSide
                    maximalCutCopy[bsub] = max(maximalCutCopy[bsub], clusterPositions[a][bsub])
                    
                for b in badSide:
                    bsub, bpos = graph.matSubPosMap[b]
                    clusterBreaksCopy[a, bsub] = badSide
                    maximalCutCopy[bsub] = max(maximalCutCopy[bsub], clusterPositions[a][bsub])
                
                nextState = (0, numOrdered, numLeft + 1, pairsLeft + pairsDiff, stateCounter, queueIdxsCopy, clusterBreaksCopy, maximalCutCopy, [], False)
                nextState = developState(nextState, graph, aggression, greedy, len(crossedClusters), subsetClusters, clusterPositions)

                nextStates.append(nextState)
                
            if greedy:
                nextState = min(nextStates, key=lambda x : x[0])
                heapq.heappush(heap, nextState)
            else:
                for nextState in nextStates:
                    heapq.heappush(heap, nextState)
                
    queueIdxs = {}
    for asub in subsetClusters:
        queueIdxs[asub] = 0
        
    orderedClusters = []
    foundGood = True
    while foundGood:
        foundGood = False
        for asub in queueIdxs:
            good = True
            idx = queueIdxs[asub]
            if idx == len(subsetClusters[asub]):
                continue
            a, pos = subsetClusters[asub][idx]
            
            if (a,asub) in clusterBreaks:
                cluster = clusterBreaks[a,asub]
            else:
                cluster = graph.clusters[a]
            
            for b in cluster:
                bsub, bpos = graph.matSubPosMap[b]
                if clusterPositions[a][bsub] != queueIdxs[bsub]:
                    good = False
                    break
                
            if good:
                orderedClusters.append(cluster)
                for b in cluster:
                    bsub, bpos = graph.matSubPosMap[b]
                    queueIdxs[bsub] = clusterPositions[a][bsub] + 1
                foundGood = True
                break
                    
    graph.clusters = orderedClusters

def developState(state, graph, aggression, greedy, crossed, subsetClusters, clusterPositions):
    heuristic, numOrdered, numLeft, pairsLeft, counter, queueIdxs, clusterBreaks, maximalCut, newClusterBreaks, safeFrontier  = state
    
    foundGood = True
    while foundGood:
        foundGood = False
        newClusterBreaks = []
        visited = set()
        safeFrontier = True
        
        for asub in queueIdxs:
            idx = queueIdxs[asub]
            if idx <= maximalCut[asub]:
                safeFrontier = False

            if idx == len(subsetClusters[asub]):
                continue
            a, pos = subsetClusters[asub][idx]
            if (a,asub) in visited:
                continue
            
            if (a,asub) in clusterBreaks:
                cluster = clusterBreaks[a,asub]
            else:
                cluster = graph.clusters[a]
            
            goodSide, badSide, crossedClusters = [], [], set()
            for b in cluster:
                bsub, bpos = graph.matSubPosMap[b]
                visited.add((a, bsub))
                bidx = clusterPositions[a][bsub]
                diff = bidx - queueIdxs[bsub]
                if diff == 0:
                    goodSide.append(b)
                else:
                    badSide.append(b)
                                   
            if len(badSide) == 0:
                for b in cluster:
                    bsub, bpos = graph.matSubPosMap[b]
                    queueIdxs[bsub] = clusterPositions[a][bsub] + 1
                numOrdered = numOrdered + 1
                numLeft = numLeft - 1
                foundGood = True
                break
            else:
                newClusterBreaks.append((a, goodSide, badSide, crossedClusters))

    
    if greedy:
        for a, goodSide, badSide, crossedClusters in newClusterBreaks:
            goodSub = set()
            for b in goodSide:
                bsub, bpos = graph.matSubPosMap[b] 
                goodSub.add(bsub)
                
            for b in badSide:    
                bsub, bpos = graph.matSubPosMap[b]   
                for i in range(queueIdxs[bsub], clusterPositions[a][bsub]):
                    c, posc = subsetClusters[bsub][i]
                    if (c,bsub) in clusterBreaks:
                        otherCluster = clusterBreaks[c,bsub]
                    else:
                        otherCluster = graph.clusters[c]
                    
                    for csite in otherCluster:
                        csub, cpos = graph.matSubPosMap[csite]
                        if csub in goodSub and clusterPositions[c][csub] > clusterPositions[a][csub]:
                            crossedClusters.add(c)
                            break  
                          
                
    if safeFrontier or len(newClusterBreaks) == 0:
        heuristic = (numLeft + numOrdered, -numOrdered, -crossed, -pairsLeft)
    else:
        heuristic = (aggression * numLeft + numOrdered, -numOrdered, -crossed, -pairsLeft)
    state = (heuristic, numOrdered, numLeft, pairsLeft, counter, queueIdxs, clusterBreaks, maximalCut, newClusterBreaks, safeFrontier)
    return state


import psutil, os
import argparse

# return memory usage of python process by MB
def memoryUsage():
    process = psutil.Process(os.getpid())
    mem = process.memory_info().rss / float(2 ** 20)
    return mem

# reformat argparse help text formatting
class SmartHelpFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if '\n' in text:
            temp = text.split('\n')
            ret = []
            for _splice in [argparse.HelpFormatter._split_lines(self, x, width)
                    for x in temp]:
                ret.extend(_splice)
            return ret
        return argparse.HelpFormatter._split_lines(self, text, width)



'''
Created on 1.22.2022 by Chengze Shen

Algorithms for tree decomposition and hmmsearch.
'''

import re, os, subprocess, math, time, psutil, shutil 
from operator import add
from functools import partial
from tqdm import tqdm

import dendropy
from dendropy.datamodel.treemodel import Tree

from witch_msa.configs import Configs, tqdm_styles
from witch_msa.gcmm import *
from witch_msa.gcmm.tree import PhylogeneticTree
from witch_msa.helpers.alignment_tools import Alignment, MutableAlignment 
from witch_msa.helpers.math_utils import lcm

import concurrent.futures

#import pyhmmer
#from helpers.pyhmmer_tools import *


'''
***** ADOPTED from sepp/exhaustive.py - ExhaustiveAlgorithm class *****
Class to perform backbone tree decomposition as does in UPP
'''
class DecompositionAlgorithm(object):
    def __init__(self, backbone_path, backbone_tree_path, alignment_size=10):
        self.symfrac = 0.0
        self.ere = 0.59
        self.informat = 'afa'
        self.molecule = Configs.molecule
        self.path = Configs.hmmbuildpath

        self.strategy = 'centroid'              # default in SEPP/UPP
        self.decomp_strategy = 'hierarchical'   # ensemble of HMMs
        self.alignment_size = alignment_size    # default: 10 as UPP
        self.minsubsetsize = 2
        self.pdistance = 1                      # default in SEPP/UPP
        self.distances = {}
        self.maxDiam = None                     # default in SEPP/UPP

        self.backbone_path = backbone_path
        self.backbone_tree_path = backbone_tree_path

        self.outdir = '{}/tree_decomp'.format(Configs.outdir)
        #self.timeout = 30

    '''
    Read in the backbone alignment and tree
    '''
    def read_alignment_and_tree(self):
        Configs.log('Reading backbone alignment: {}'.format(
            self.backbone_path))
        assert os.path.exists(self.backbone_path), \
                'Backbone alignment [{}] does not exist!'.format(
                        self.backbone_path)
        alignment = Alignment()
        alignment.read_file_object(self.backbone_path)

        Configs.log('Reading backbone tree: {}'.format(
            self.backbone_tree_path))
        assert os.path.exists(self.backbone_tree_path), \
                'Backbone tree [{}] does not exist'.format(
                        self.backbone_tree_path)
        tree = PhylogeneticTree(dendropy.Tree.get_from_stream(
            open(self.backbone_tree_path, 'r'),
            schema='newick',
            preserve_underscores=True))
        
        return alignment, tree

    '''
    Do the decomposition on the backbone alignment/tree
    Take in a ProcessPoolExecutor for parallelism
    '''
    def decomposition(self, lock, pool):
        start = time.time()
        Configs.log('Started decomposing the backbone to eHMM')
        alignment, tree = self.read_alignment_and_tree()

        assert isinstance(alignment, Alignment)
        assert isinstance(tree, PhylogeneticTree)

        tree.get_tree().resolve_polytomies()
        
        # label edges with numbers so that we can assemble them back
        # at the end
        tree.label_edges()

        # decompose the tree into alignment subsets
        alignment_tree_map = PhylogeneticTree(
                Tree(tree.den_tree)).decompose_tree(
                        self.alignment_size,
                        strategy=self.strategy,
                        minSize=self.minsubsetsize,
                        tree_map={},
                        decomp_strategy=self.decomp_strategy,
                        pdistance=self.pdistance,
                        distances=self.distances,
                        maxDiam=self.maxDiam)
        assert len(alignment_tree_map) > 0, (
                'Tree could not be decomposed '
                'given the following settings: '
                'strategy: {}\nminsubsetsize: {}\nalignment_size: {}'.format(
                    self.strategy, self.minsubsetsize, self.alignment_size))
        
        Configs.debug('Alignment subsets: {}'.format(len(alignment_tree_map)))

        subset_args = []
        for a_key, a_tree in alignment_tree_map.items():
            assert isinstance(a_tree, PhylogeneticTree)
            label = 'A_0_{}'.format(a_key)
            subset_taxa = a_tree.leaf_node_names()
            subset_args.append((label, subset_taxa))
        
        Configs.log('Creating an ensemble of HMMs: {} subsets'.format(
            len(subset_args)))

        # create all subset alignments and HMMBuild them
        if self.molecule is None:
            self.molecule = Configs.inferDataType(self.backbone_path)
        outdirprefix = self.outdir + '/root'
        func = partial(subset_alignment_and_hmmbuild, lock, 
                self.path, outdirprefix,
                self.molecule, self.ere, self.symfrac,
                self.informat, self.backbone_path)
        ret = list(pool.map(func, subset_args))
        hmmbuild_paths = []
        
        # record the retained columns in each HMM subset
        # (used in mapping hmmalign results to alignment graph)
        subset_to_retained_columns = dict()
        subset_to_nongaps_per_column = dict()
        for item in ret:
            hmmbuild_paths.append(item[0])
            # item[1] -- A_0_*
            ind = int(item[1].split('_')[-1])
            subset_to_retained_columns[ind] = item[2]
            subset_to_nongaps_per_column[ind] = item[3]

        # convert to list
        subset_to_retained_columns = [subset_to_retained_columns[k]
                    for k in sorted(subset_to_retained_columns.keys())]
        subset_to_nongaps_per_column = [subset_to_nongaps_per_column[k]
                    for k in sorted(subset_to_nongaps_per_column.keys())]
        
        assert len(hmmbuild_paths) == len(subset_args), \
                'Number of HMMs created does not match ' \
                'the original number of subsets'
        Configs.log('Finished creating {} HMMs to {}'.format(
            len(hmmbuild_paths), outdirprefix))
        
        dur = time.time() - start
        Configs.runtime(' '.join(['(DecompositionAlgorithm.decomposition)',
                'Time to decompose the backbone (s):', str(dur)]))
        return hmmbuild_paths, subset_to_retained_columns, \
                subset_to_nongaps_per_column

'''
Class to perform HMMSearch on all hmmbuild subsets and fragment sequences
(Also will break fragments to fragment chunks for better parallelism)
'''
class SearchAlgorithm(object):
    def __init__(self, hmmbuild_paths):
        self.hmmbuild_paths = hmmbuild_paths
        self.unaligned = None
        self.path = Configs.hmmsearchpath

        self.max_chunk_size = 20000             # default in SEPP/UPP
        self.molecule = Configs.molecule

        self.filters = False                    # No filters
        self.elim = 99999999                    # elim value for hmmsearch
        self.piped = False                      # default in SEPP/UPP
        
        self.outdir = Configs.outdir + '/tree_decomp'

    ####### ONLY USED WHEN PRESENTS WITH AN HMM DIRECTORY #######
    def readHMMDirectory(self, lock, pool):
        subset_to_retained_columns = dict()
        subset_to_nongaps_per_column = dict()
        
        # use "find" command to find all subset directories
        cmd = 'find {} -maxdepth 2 -name A_0_* -type d'.format(Configs.hmmdir)
        subset_dirs = [os.path.realpath(x) 
                for x in os.popen(cmd).read().split('\n')[:-1]]
        Configs.log('Found existing HMM directory: {}'.format(Configs.hmmdir))
        Configs.log('Reading {} subsets...'.format(len(subset_dirs)))
        
        # terminate if not finding any HMMs in the current directory
        if len(subset_dirs) == 0:
            msg = 'Cannot find any pre-existing HMMs in {}!'.format(
                    Configs.hmmdir) + ' Please remove the directory and rerun.'
            Configs.error(msg)
            raise FileNotFoundError(msg)
                    
        # use pool to process all subset dirs
        subset_args = []
        tmp_map = dict()
        for sdir in subset_dirs:
            ind = int(sdir.split('/')[-1].split('_')[-1])
            subset_args.append((ind, sdir))
            tmp_map[ind] = sdir

        # obtain the backbone alignment, which should be of A_0_0
        cmd = 'find {} -maxdepth 1 -name hmmbuild.input.*fasta -type f'.format(
                tmp_map[0])
        bb_path = os.path.realpath(os.popen(cmd).read().split('\n')[0])
        bb_aln = Alignment(); bb_aln.read_file_object(bb_path)
        
        func = partial(subset_obtain_retained_columns, bb_aln)
        ret = list(pool.map(func, subset_args))

        for item in ret:
            subset_to_retained_columns[int(item[0])] = item[1]
            subset_to_nongaps_per_column[int(item[0])] = item[2]

        # convert to list
        subset_to_retained_columns = [subset_to_retained_columns[k]
                    for k in sorted(subset_to_retained_columns.keys())]
        subset_to_nongaps_per_column = [subset_to_nongaps_per_column[k]
                    for k in sorted(subset_to_nongaps_per_column.keys())]

        #del bb_aln
        Configs.log('Finished reading decomposition subsets...')
        return bb_path, subset_to_retained_columns, subset_to_nongaps_per_column
    
    # main function to perform all-against-all query-HMM searches in MP 
    def search(self, lock, pool):
        Configs.log('Running all-against-all HMM searches between queries ' \
                'and HMMs...')
        start = time.time()
        # create fragment chunks and run HMMSearch on all subsets
        # **** USING SEPP exhaustive implementation to find the best number
        # **** of chunks (Least Common Multiple of #subsets & #cpus // #subsets)
        num_chunks = lcm(
                len(self.hmmbuild_paths),
                Configs.num_cpus) // len(self.hmmbuild_paths)
        #num_chunks = Configs.num_cpus
        frag_chunk_paths = self.read_and_divide_unaligned(num_chunks)

        # create subproblems of HMMSearch
        subset_args = []
        for hmmbuild_path in self.hmmbuild_paths:
            hmmsearch_outdir = '/'.join(hmmbuild_path.split('/')[:-1])
            hmm_label = hmmbuild_path.split('/')[-1].split('.')[-1]
            for i in range(0, len(frag_chunk_paths)):
                frag_chunk_path = frag_chunk_paths[i]
                subset_args.append((hmmsearch_outdir, hmm_label,
                    hmmbuild_path, frag_chunk_path, i))

        # run each subproblem
        if self.molecule is None:
            self.molecule = Configs.inferDataType(self.backbone_path)
        func = partial(subset_frag_chunk_hmmsearch, lock, self.path,
                self.piped, self.elim, self.filters, self.molecule)
        hmmsearch_paths, futures = [], []
        for subset_arg in subset_args:
            futures.append(pool.submit(func, subset_arg))
        for future in tqdm(
                concurrent.futures.as_completed(futures),
                total=len(subset_args), **tqdm_styles):
            res = future.result()
            if res:
                hmmsearch_paths.append(res)
        assert len(hmmsearch_paths) == len(subset_args), \
                'It seems that some HMMSearch jobs failed'
        Configs.log('Finished {} HMMSearch jobs.'.format(len(hmmsearch_paths)))

        dur = time.time() - start
        Configs.runtime(' '.join(['(SearchAlgorithm.search) Time to run',
                'all-against-all HMMSearches (s):', str(dur)]))
        return frag_chunk_paths 

    def read_and_divide_unaligned(self, num_chunks, extra_frags={}):
        max_chunk_size = self.max_chunk_size
        Configs.log('Reading in fragment files and breaking ' + 
                'them into at most {} chunks '.format(num_chunks) +
                '(each having at most {} sequences)'.format(max_chunk_size))

        # read in query paths (either provided or produced by separating
        # original input sequences to backbone and query sequences)
        unaligned = MutableAlignment()
        unaligned.read_file_object(Configs.query_path)
        ids_unaligned = unaligned.keys()

        # test if input fragment names contain whitespaces or tabs
        ids_unaligned_violations = [id_ for id_ in ids_unaligned
                if (' ' in id_) or ('\t' in id_)]
        if len(ids_unaligned_violations) > 0:
            raise ValueError(
                "Your input fragment file contains {} sequences, ".format(
                    len(ids_unaligned_violations)) + 
                "which names contain either whitespaces or tabs '\\t'. " +
                "Their names are:\n {}".format(
                    "'\n'  ".join(ids_unaligned_violations)))

        # add in extra fragments if provided
        for k, v in extra_frags.items():
            unaligned[k] = v.replace("-", "")

        alg_chunks = unaligned.divide_to_equal_chunks(num_chunks,
                max_chunk_size)
        
        # write to local (<WITCH_OUTPUT_DIR>/tree_decomp/fragment_chunks)
        # and save all frag chunk paths
        ret = []
        fc_outdir = self.outdir + '/fragment_chunks'
        if not os.path.isdir(fc_outdir):
            os.makedirs(fc_outdir)

        for i in range(0, len(alg_chunks)):
            temp_file = None
            if alg_chunks[i]:
                temp_file = '{}/fragment_chunk_{}.fasta'.format(fc_outdir, i)
                alg_chunks[i].write(temp_file, 'FASTA')
                Configs.debug('Writing alignment chunk #{} to {}'.format(
                    i, temp_file))
                ret.append(temp_file)
        Configs.log("Finished breaking fragments into {} chunks to {}".format(
            len(ret), fc_outdir))
        return ret


################ HELPER FUNCTIONS FOR MULTI-PROCESSING ################

'''
Obtain subset alignment given taxa, and run hmmbuild on the subset
alignment.
'''
def subset_alignment_and_hmmbuild(lock, binary, outdirprefix, molecule, 
        ere, symfrac, informat, backbone_path, args):
    label, taxa = args
    alignment = Alignment(); alignment.read_file_object(backbone_path)
    subalignment = alignment.sub_alignment(taxa)
    retained_columns = subalignment.delete_all_gaps()
    
    # also accumulate the number of nongaps for each column
    nongaps_per_column = [0] * subalignment.sequence_length() 
    for seq in subalignment.values():
        nongaps_per_column = map(add, nongaps_per_column,
                [int(x != '-') for x in seq])
    del alignment

    outdir = os.path.join(outdirprefix, label)
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    
    # write subalignment to outdir
    subalignment_path = '{}/hmmbuild.input.{}.fasta'.format(outdir, label)
    subalignment.write(subalignment_path, 'FASTA')

    ############# 9.19.2023 ############
    ## added version to use pyhmmer for building
    #hmmbuild_path = '{}/hmmbuild.model.{}'.format(outdir, label)
    #alphabet = moleculeToAlphabet(molecule)
    #msa_d = alignmentToTextMSA(
    #        subalignment, name=f'hmmbuild.model.{label}').digitize(alphabet)

    #builder = pyhmmer.plan7.Builder(alphabet, ere=ere, symfrac=symfrac)
    #background = pyhmmer.plan7.Background(alphabet)
    #hmm, _, _ = builder.build_msa(msa_d, background)
    #with open(hmmbuild_path, 'wb') as fh:
    #    hmm.write(fh, binary=False)
    #lock.acquire()
    #try:
    #    Configs.debug('[pyhmmer.plan7.Builder] ere={}, symfrac={}'.format(
    #        ere, symfrac))
    #finally:
    #    lock.release()
    #    return (hmmbuild_path, label, tuple(retained_columns),
    #            tuple(nongaps_per_column))

    # run HMMBuild with 1 cpu given the subalignment
    hmmbuild_path = '{}/hmmbuild.model.{}'.format(outdir, label)
    cmd = [binary, '--cpu', '1',
            '--{}'.format(molecule),
            '--ere', str(ere),
            '--symfrac', str(symfrac),
            '--informat', informat,
            '-o', '/dev/null']
    cmd.extend([hmmbuild_path, subalignment_path])
    os.system(' '.join(cmd))
    lock.acquire()
    try:
        Configs.debug('[HMMBuild] Command used: {}'.format(' '.join(cmd)))
    finally:
        lock.release()
        return (hmmbuild_path, label, tuple(retained_columns),
                tuple(nongaps_per_column))

'''
a single HMMSearch job between a frag chunk and an hmm
'''
def subset_frag_chunk_hmmsearch(lock, binary, piped, elim, filters,
        molecule, args):
    outdir, hmm_label, hmmbuild_path, unaligned, frag_index = args
    
    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    ########### 9.19.2023 ##############
    ## use pyhmmer.plan7.Pipeline to do the HMMSearch
    #hmmsearch_path = '{}/hmmsearch.results.{}.fragment_chunk_{}'.format(
    #        outdir, hmm_label, frag_index)
    #alphabet = moleculeToAlphabet(molecule)
    #background = pyhmmer.plan7.Background(alphabet)

    ## load HMM
    #hmm = None
    #with pyhmmer.plan7.HMMFile(hmmbuild_path) as hmm_file:
    #    hmm = hmm_file.read()
    #
    ## set up the pipeline
    #kwargs = {'E': elim}
    #if not filters:
    #    kwargs.update({'bias_filter': False, 'F1': 1.0, 'F2': 1.0, 'F3': 1.0})
    #pipeline = pyhmmer.plan7.Pipeline(alphabet, background=background,
    #        **kwargs)

    ## run the pipeline
    #tophits = None
    #with pyhmmer.easel.SequenceFile(unaligned,
    #        digital=True, alphabet=alphabet) as seq_file:
    #    tophits = pipeline.search_hmm(hmm, seq_file)
    #res = evalHMMSearchOutputPyhmmer(tophits)
    #with open(hmmsearch_path, 'w') as f:
    #    f.write(str(res))
    #lock.acquire()
    #try:
    #    Configs.debug('[pyhmmer.plan7.Pipeline.search_hmm] query={}, model={}, outpath={}, kwargs={}'.format(
    #        unaligned, hmmbuild_path, hmmsearch_path, kwargs))
    #finally:
    #    lock.release()
    #    return hmmsearch_path

    hmmsearch_path = '{}/hmmsearch.results.{}.fragment_chunk_{}'.format(
            outdir, hmm_label, frag_index)
    cmd = [binary, '--cpu', '1', '--noali', '-E', str(elim)]
    if not piped:
        cmd.extend(['-o', hmmsearch_path])
    if not filters:
        cmd.extend(['--max'])
    cmd.extend([hmmbuild_path, unaligned])
    os.system(' '.join(cmd))

    # modify the output file and only retain taxon name, E-value, bit-score
    res = evalHMMSearchOutput(hmmsearch_path)
    with open(hmmsearch_path, 'w') as f:
        f.write(str(res))

    lock.acquire()
    try:
        Configs.debug('[HMMSearch] Command used: {}'.format(' '.join(cmd)))
    finally:
        lock.release()
        return hmmsearch_path

'''
helper function to read alignment from a subset alignment and obtain
its retained columns with respect to the backbone alignment
Additionally, return the number of non-gap characters for each retained column
'''
def subset_obtain_retained_columns(bbaln, args):
    ind, sdir = args
    # find the subset alignment path
    cmd = 'find {} -maxdepth 1 -name hmmbuild.input.*fasta -type f'.format(
            sdir)
    try:
        path = os.path.realpath(os.popen(cmd).read().split('\n')[0])
        aln = Alignment(); aln.read_file_object(path)
        keys = aln.keys()
        subaln = bbaln.sub_alignment(keys)
        retained_columns = subaln.delete_all_gaps()
        
        # accumulate the number of nongaps for each column
        nongaps_per_column = [0] * subaln.sequence_length() 
        for seq in subaln.values():
            nongaps_per_column = map(add, nongaps_per_column,
                    [int(x != '-') for x in seq])

        return ind, tuple(retained_columns), tuple(nongaps_per_column)
    except FileNotFoundError as e:
        # cannot find the hmm input path in the corresponding directory
        raise Exception(
            'Cannot load subset {} at {}! Decomposition may be incomplete.'.format(
                ind, sdir))

'''
helper function for modifying HMMSearch output file
'''
def evalHMMSearchOutput(path):
    outfile = open(path, 'r')
    results = {}

    pattern = re.compile(
        r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+"
        r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)")
    start_reading = False
    for line in outfile:
        line = line.strip()
        if not start_reading and line.startswith("E-value") is True:
            start_reading = True
        elif start_reading and line == "":
            start_reading = False
            break
        elif start_reading:
            matches = pattern.search(line)
            if matches is not None and matches.group(0).find("--") == -1:
                results[matches.group(9).strip()] = (
                    float(matches.group(1).strip()),
                    float(matches.group(2).strip()))
                # _LOG.debug("Fragment scores;"
                #           "fragment:%s E-Value:%s BitScore:%s" %(matches
                # .group(9).strip(),matches.group(1).strip(), matches.
                # group(2).strip()))
    outfile.close()
    return results

'''
helper function to keep just taxon name, e-value, and bit-score from tophits
'''
def evalHMMSearchOutputPyhmmer(tophits):
    results = {}
    for hit in tophits:
        name, evalue, score = hit.name.decode(), hit.evalue, hit.score
        results[name] = (evalue, f'{score:.1f}')
    return results



'''
Created on Apr 14, 2020

@author: Vlad
'''

import os
import time
import random

from align.merge.alignment_graph import AlignmentGraph
from helpers import sequenceutils, hmmutils
from tasks import task
from configuration import Configs
from tools import external_tools

'''
Building a MAGUS alignment graph from backbone alignments. 
The graph is a sparse matrix, stored as a weighted adjacency list.
Backbone alignment tasks are run in parallel, and can begin before the subalignments are finished.
When subalignments finish, we can initialize the alignment graph. 
Backbones are added to this graph as they complete.
Graph is written to file in MCL-compliant format.
'''

def buildGraph(context):
    time1 = time.time() 
    
    context.graph = AlignmentGraph(context)
    context.initializeSequences()
    
    if os.path.exists(context.graph.graphPath):
        Configs.log("Found existing graph file {}".format(context.graph.graphPath))
    else:
        requestBackboneTasks(context)
    
    context.awaitSubalignments()
    context.graph.initializeMatrix()
    
    if os.path.exists(context.graph.graphPath):
        context.graph.readGraphFromFile(context.graph.graphPath)
    else:
        context.initializeBackboneSequenceMapping()
        '''
            6.1.2021 - added by Chengze Shen
            a new invocation to the newly added function in context
            which initializes the backbone weights
        '''
        context.initializeBackboneWeights()

        buildMatrix(context)
        context.graph.writeGraphToFile(context.graph.graphPath)
       
    time2 = time.time()
    Configs.log("Built the alignment graph in {} sec..".format(time2-time1))

def requestBackboneTasks(context):
    if len(context.backbonePaths) > 0:
        Configs.log("Using {} user-defined backbone files..".format(len(context.backbonePaths)))
        context.backbonePaths = context.backbonePaths
        for path in context.backbonePaths:
            context.backboneTaxa.update(sequenceutils.readFromFasta(path))
    
    elif Configs.graphBuildMethod == "mafft":
        Configs.log("Using {} MAFFT backbones..".format(Configs.mafftRuns)) 
        requestMafftBackbones(context)
        
    elif Configs.graphBuildMethod == "subsethmm":
        Configs.log("Using {} HMM-extended subalignments as backbone files..".format(len(context.subalignmentPaths)))
        context.backbonePaths = context.subalignmentPaths
        context.backboneExtend.update(context.backbonePaths)
        
    elif Configs.graphBuildMethod == "initial":
        Configs.log("Using the initial decomposition alignment as the single backbone..")
        initialAlignPath = os.path.join(context.workingDir, "decomposition", "initial_tree", "initial_insert_align.txt")
        context.backbonePaths = [initialAlignPath]
    
    if not Configs.constrain and Configs.graphBuildMethod != "subsethmm":
        context.backbonePaths.extend(context.subalignmentPaths)

def requestMafftBackbones(context):
    numTaxa = max(1, int(Configs.mafftSize/len(context.subsetPaths)))

    for n in range(Configs.mafftRuns):
        unalignedFile = os.path.join(context.graph.workingDir, "backbone_{}_unalign.txt".format(n+1))
        alignedFile = os.path.join(context.graph.workingDir, "backbone_{}_mafft.txt".format(n+1))
        if os.path.exists(alignedFile):
            Configs.log("Existing backbone file found: {}".format(alignedFile))
            backbone = sequenceutils.readFromFasta(alignedFile)
            context.backbonePaths.append(alignedFile)
        else:
            backbone = assignBackboneTaxa(context, numTaxa, unalignedFile)
            backboneTask = external_tools.buildMafftAlignment(unalignedFile, alignedFile)
            context.backboneTasks.append(backboneTask)
            
        if Configs.graphBuildHmmExtend:
            context.backboneExtend.add(alignedFile)
        else:
            context.backboneTaxa.update(backbone)
    task.submitTasks(context.backboneTasks)    
                
def buildMatrix(context):
    addedBackbones = set()
    for backboneTask in task.asCompleted(context.backboneTasks):
        addAlignmentFileToGraph(context, backboneTask.outputFile)
        addedBackbones.add(backboneTask.outputFile)
    
    for backboneFile in context.backbonePaths:
        if backboneFile not in addedBackbones:
            addAlignmentFileToGraph(context, backboneFile)  
    
def assignBackboneTaxa(context, numTaxa, unalignedFile):
    backbone = {}
    for subset in context.subsets:
        random.shuffle(subset)
        for taxon in subset[:numTaxa]:
            backbone[taxon] = context.unalignedSequences[taxon]
    sequenceutils.writeFasta(backbone, unalignedFile)
    return backbone
    
def addAlignmentFileToGraph(context, alignedFile):
    Configs.log("Feeding backbone {} to the graph..".format(alignedFile))
    backboneAlign = sequenceutils.readFromFasta(alignedFile)  
    alignmentLength = len(next(iter(backboneAlign.values())).seq)
    
    if alignedFile in context.backboneExtend:
        extensionTasks = requestHmmExtensionTasks(context, backboneAlign, alignedFile)
        task.submitTasks(extensionTasks)
        for extensionTask in task.asCompleted(extensionTasks):
            backboneAlign.update(sequenceutils.readFromStockholm(extensionTask.outputFile, includeInsertions=True))
    
    alignmap = backboneToAlignMap(context, backboneAlign, alignmentLength)
    Configs.log("Constructed backbone alignment map from {}".format(alignedFile))
    
    '''
        6.1.2021 - added by Chengze Shen
        read in the weight for the alignedFile if the weight exists
    '''
    weight = 1
    if context.backboneWeightsPath:
        # the weight has to be defined for this to work
        weight = context.backboneWeights.get(alignedFile)

    graph = context.graph  
    with graph.matrixLock:
        for l in range(alignmentLength):  
            for a, avalue in alignmap[l].items():
                for b, bvalue in alignmap[l].items():
                    
                    if Configs.graphBuildRestrict:
                        asub, apos = graph.matSubPosMap[a] 
                        bsub, bpos = graph.matSubPosMap[b] 
                        if asub == bsub and apos != bpos:
                            continue

                    '''
                        6.1.2021 - modified by Chengze Shen
                        instead of a plain number of counts, we consider
                        a user-defined weighting for the backbone
                        
                        ** the assumption is that the backbone weighting is
                        ** tied to only one query sequence (because if we 
                        ** have multiple queries in the backbone alignment,
                        ** the weightings of the bb of queries may not be
                        ** the same.
                    '''
                    #graph.matrix[a][b] = graph.matrix[a].get(b,0) + avalue * bvalue         
                    graph.matrix[a][b] = graph.matrix[a].get(b,0) \
                            + avalue * bvalue * weight
    '''
        7.8.2021 - added by Chengze Shen
        log the weight information for the backbone too
    '''
    Configs.log("Fed backbone {} to the graph - weight = {}.".format(
        alignedFile, weight))
    #Configs.log("Fed backbone {} to the graph.".format(alignedFile))

def backboneToAlignMap(context, backboneAlign, alignmentLength):
    alignmap = [{} for i in range(alignmentLength)]
    t = 0
    
    for taxon in backboneAlign:
        subsetIdx = context.taxonSubalignmentMap[taxon]
        subsetseq = context.backboneSubalignment[taxon].seq
        unalignedseq = context.unalignedSequences[taxon].seq     
        backboneseq = backboneAlign[taxon].seq   
        
        i = 0
        posarray = [0] * len(unalignedseq)
        for n in range(len(subsetseq)):
            if subsetseq[n] == unalignedseq[i]:
                posarray[i] = n 
                i = i + 1
                if i == len(unalignedseq):
                    break
        
        i = 0
        n = 0
        for c in backboneseq:
            if i == len(unalignedseq):
                break
            if c == unalignedseq[i]:
                position = int(context.graph.subsetMatrixIdx[subsetIdx] + posarray[i])
                alignmap[n][position] = alignmap[n].get(position, 0) + 1
            if c.upper() == unalignedseq[i]:
                i = i + 1
            if c == c.upper() and c != '.':
                n = n + 1
                
        t = t + 1
            
    return alignmap
        
def requestHmmExtensionTasks(context, backbone, alignedFile):
    baseName = os.path.basename(alignedFile)
    hmmDir = os.path.join(context.graph.workingDir, "hmm_{}".format(baseName))
    extensionUnalignedFile = os.path.join(hmmDir, "queries.txt")
    hmmPath = os.path.join(hmmDir, "hmm_model.txt")
    if not os.path.exists(hmmDir):
        os.makedirs(hmmDir)
    
    backboneExtension = {}
    for taxon in context.unalignedSequences:
        if not taxon in backbone:
            backboneExtension[taxon] = context.unalignedSequences[taxon]
                
    sequenceutils.writeFasta(backboneExtension, extensionUnalignedFile) 
    buildTask = hmmutils.buildHmmOverAlignment(alignedFile, hmmPath)
    buildTask.run()
    alignTasks = hmmutils.hmmAlignQueries(hmmPath, extensionUnalignedFile)
    return alignTasks 
    



'''
Created on Apr 14, 2020

@author: Vlad
'''

import subprocess
import os
import random
import shutil
from configuration import Configs
from tasks.task import Task

def runCommand(**kwargs):
    command = kwargs["command"]
    Configs.log("Running an external tool, command: {}".format(command))
    runner = subprocess.run(command, shell = True, cwd = kwargs["workingDir"], universal_newlines = True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:    
        runner.check_returncode()
    except:
        Configs.error("Command encountered error: {}".format(command))
        Configs.error("Exit code: {}".format(runner.returncode))
        Configs.error("Output: {}".format(runner.stdout))
        raise
    for srcPath, destPath in kwargs.get("fileCopyMap", {}).items():
        shutil.move(srcPath, destPath)

def runClustalOmegaGuideTree(fastaPath, workingDir, outputPath, threads = 1):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.clustalPath]
    args.extend(["-i", fastaPath, "--max-hmm-iterations=-1", "--guidetree-out={}".format(tempPath)])
    args.extend(["--threads={}".format(threads)])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def generateMafftFilePathMap(inputPaths, outputDir):
    mafftMap = {inputPath : os.path.join(outputDir, "mafft_{}".format(os.path.basename(inputPath))) for inputPath in inputPaths}
    return mafftMap

def buildMafftAlignments(inputOutputPathMap):
    tasks = [buildMafftAlignment(inputPath, outputPath) for inputPath, outputPath in inputOutputPathMap.items()]
    return tasks
    
def buildMafftAlignment(inputPath, outputPath, subtablePath = None):
    return runMafft(inputPath, subtablePath, Configs.workingDir, outputPath, Configs.numCores)                

def runMafft(fastaPath, subtablePath, workingDir, outputPath, threads = 1):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.mafftPath, "--localpair", "--maxiterate", "1000", "--ep", "0.123", 
            "--quiet", "--thread", str(threads), "--anysymbol"]
    if subtablePath is not None:
        args.extend(["--merge", subtablePath])
    args.extend([fastaPath, ">", tempPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runMafftGuideTree(fastaPath, workingDir, outputPath, threads = 1):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    treeFile = os.path.join(os.path.dirname(fastaPath),  "{}.tree".format(os.path.basename(fastaPath)))
    args = [Configs.mafftPath, "--retree", "0", "--treeout", "--parttree",
            "--quiet", "--thread", str(threads), "--anysymbol"]
    args.extend(["--partsize", "1000"])
    args.extend([fastaPath, ">", tempPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {treeFile : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runMcl(matrixPath, inflation, workingDir, outputPath):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.mclPath, matrixPath, "--abc", "-o", tempPath]
    if inflation is not None:
        args.extend(["-I", str(inflation)])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runMlrMcl(matrixPath, granularity, balance, inflation, workingDir, outputPath):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.mlrmclPath, matrixPath, "-o", tempPath]
    if granularity is not None:
        args.extend(["-c", str(granularity)])
    if balance is not None:
        args.extend(["-b", str(balance)])    
    if inflation is not None:
        args.extend(["-i", str(inflation)])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runFastTree(fastaFilePath, workingDir, outputPath, mode = "normal", intree = None):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    
    args = [Configs.fasttreePath]
    if Configs.inferDataType(fastaFilePath) == "protein":
        args.extend(["-lg"])
    else:
        args.extend(["-nt", "-gtr"])
    
    if intree is not None:
        args.extend(["-intree", intree])
    
    if mode == "fast":
        args.extend(["-fastest", "-nosupport"]) 
    elif mode == "faster":
        args.extend(["-fastest", "-nosupport", "-mlnni", "4" ]) 
    elif mode == "noml":
        args.extend(["-fastest", "-nosupport", "-noml"])
    
    args.extend([fastaFilePath, ">", tempPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runRaxmlNg(fastaFilePath, workingDir, outputPath, threads = 8):
    # raxml-ng --msa prim.phy --model GTR+G --prefix T4 --threads 2 --seed 2 --tree pars{25},rand{25}
    baseName = os.path.basename(outputPath).replace(".","")
    raxmlFile = os.path.join(workingDir, "{}.raxml.bestTree".format(baseName))
    seed = random.randint(1, 1000000)
    args = [Configs.raxmlPath,
            "--msa", fastaFilePath,
            "--prefix", baseName,
            "--threads", str(threads),
            "--seed", str(seed)]
    
    if Configs.inferDataType(fastaFilePath) == "protein":
        args.extend(["--model", "LG+G"])
    else:
        args.extend(["--model", "GTR+G"])
        
    args.extend(["--tree", "pars{{{}}}".format(1)])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {raxmlFile : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runHmmBuild(alignmentPath, workingDir, outputPath):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.hmmbuildPath,'--ere', '0.59', "--cpu", "1"]
    args.extend(["--symfrac", "0.0", "--informat", "afa", tempPath, alignmentPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runHmmAlign(hmmModelPath, fragPath, workingDir, outputPath):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.hmmalignPath, "-o", tempPath]
    args.extend([hmmModelPath, fragPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)

def runHmmSearch(hmmModelPath, fragPath, workingDir, outputPath):
    tempPath = os.path.join(os.path.dirname(outputPath), "temp_{}".format(os.path.basename(outputPath)))
    args = [Configs.hmmsearchPath,"--noali", "--cpu", "1", "-o", tempPath, "-E", "99999999", "--max"]
    args.extend([hmmModelPath, fragPath])
    taskArgs = {"command" : subprocess.list2cmdline(args), "fileCopyMap" : {tempPath : outputPath}, "workingDir" : workingDir}
    return Task(taskType = "runCommand", outputFile = outputPath, taskArgs = taskArgs)


from witch_msa.configs import Configs
from witch_msa.helpers.alignment_tools import ExtendedAlignment
import gzip

'''
Callback for a query alignment result
*args should have three fields: _query, _index, checkpoint_path
'''
def callback_queryAlignment(success, ignored, retry, i_retry, 
        query, index, taxon_name, checkpoint_path):
    if (not query) and i_retry > 0:
        retry.append(index)
    else:
        # failed job indicated in the <witch-ng> or <default> pipelines,
        # should be ignored in the output
        if (not query) or len(query) == 0:
            ignored.append(taxon_name)
        else:
            # write to checkpoint_path
            if (not isinstance(query, ExtendedAlignment)) or (len(query) != 1):
                return
            seq = query[taxon_name]
            line = '{}\t{}\n'.format(taxon_name, seq)
            encoded = line.encode('utf-8')
            with gzip.open(checkpoint_path, 'ab') as f:
                f.write(encoded)

            # update success
            success.append(query)



'''
Created on Jun 15, 2020

@author: Vlad
'''

import heapq
import time 
import random

from configuration import Configs
    
'''
Fiduccia-Mattheyses implementation for clustering and/or tracing. 
Currently not recommended, slower and less accurate than minclusters or mwtgreedy.
'''

def fmAlgorithm(graph):
    Configs.log("Finding graph trace with FM Algorithm..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)]  
    
    if graph.clusters is None or len(graph.clusters) == 0:
        graph.buildNodeEdgeDataStructure()
    else:
        graph.buildNodeEdgeDataStructureFromClusters()
    clusters, totalCost, cuts = fmPartition(graph, lowerBound, upperBound)
    
    graph.clusters = clusters
                    
def fmPartitionWithCuts(graph, lowerBound, upperBound, cuts):
    allCuts = [lowerBound] + cuts + [upperBound]
    clusters, cost, cuts = [], 0, []
    for i in range(len(allCuts)-1):
        newClusters, newCost, newCuts = fmPartition(graph, allCuts[i], allCuts[i+1], False)
        clusters = clusters + newClusters
        cost = cost + newCost
        cuts = cuts + newCuts    
    return clusters, cost, cuts
    
def fmPartition(graph, lowerBound, upperBound, iterate = True):
    k = len(graph.context.subalignments)
    
    finished = True
    cluster = []
    for i in range(k):
        if upperBound[i] - lowerBound[i] > 1:
            finished = False
            break
        elif upperBound[i] - lowerBound[i] == 1:
            cluster.append(lowerBound[i])
    if finished:
        return [cluster], 0, []   
    
    startingCut = [int((lowerBound[i] + upperBound[i])*0.5) for i in range(k)]    
    bestCut, bestCutCost = fmFindBestCut(graph, lowerBound, upperBound, startingCut, None)    
        
    lowerClusters, lowerCost, lowerCuts = fmPartition(graph, lowerBound, bestCut)
    upperClusters, upperCost, upperCuts = fmPartition(graph, bestCut, upperBound)
    clusters = lowerClusters + upperClusters
    totalCost = bestCutCost + lowerCost + upperCost
    totalCuts = lowerCuts + [bestCut] + upperCuts

    return clusters, totalCost, totalCuts

def fmFindBestCut(graph, lowerBound, upperBound, startingCut, widthSumLimit):
    #Configs.log("Finding FM partition..")
    #Configs.log("    Lower bound: {}".format(graph.cutString(lowerBound)))
    #Configs.log("    Upper bound: {}".format(graph.cutString(upperBound)))
    #Configs.log("    Startng cut: {}".format(graph.cutString(startingCut)))
    k = len(graph.context.subalignments)
    bestCut = startingCut
    bestCutGains, bestCutCost = populateGains(graph, lowerBound, upperBound, bestCut)
    
    oldBestCut = None
    while bestCut != oldBestCut:

        if bestCutCost == 0:
            break
        
        oldBestCut = bestCut
        cutCost = bestCutCost
        cut = list(bestCut)
        gains = dict(bestCutGains)
        heap = []
        heapGains = {}
        heapGainsVersions = {}
        locked = set()
        lowerUpdateBound = [cut[i] for i in range(k)]
        upperUpdateBound = [cut[i] for i in range(k)]
        
        while True:
            newLowerBound, newUpperBound = findNewBounds(graph, lowerBound, upperBound, cut, widthSumLimit)            
            updateList = getHeapGainUpdateList(graph, newLowerBound, newUpperBound, lowerUpdateBound, upperUpdateBound)   
            updateHeapGainList(graph, cut, updateList, gains, heapGains, heapGainsVersions, heap, locked) 
            
            found = False
            reinsert = []
            while len(heap) > 0:
                gain, node, gainVersion = heapq.heappop(heap)
                if node in locked or gainVersion != heapGainsVersions[node]:
                    continue
                
                asub, apos = graph.matSubPosMap[node]
                if node < newLowerBound[asub] or node >= newUpperBound[asub]:
                    reinsert.append((gain, node, gainVersion))
                    continue

                found = True
                break
            
            if not found:
                break
            for item in reinsert:
                heapq.heappush(heap, item)
            
            locked.add(node)
            gain = gain * -1
            #asub, apos = graph.matSubPosMap[node]
            movedNodes = []
            if node >= cut[asub]:
                movedNodes = [j for j in range(cut[asub], node+1)]
                cut[asub] = node+1
            else:
                movedNodes = [j for j in range(node, cut[asub])]
                cut[asub] = node
            
            updateGains(graph, lowerBound, upperBound, cut, gains, movedNodes, lowerUpdateBound, upperUpdateBound) 
            
            cutCost = cutCost - gain
            if cutCost < bestCutCost:
                bestCut = list(cut)
                bestCutGains = dict(gains)
                bestCutCost = cutCost
                
    Configs.log("Found FM partition {}".format(graph.cutString(bestCut)))
    Configs.log("    Partition cost: {}".format(bestCutCost))  
    return bestCut, bestCutCost

def populateGains(graph, lowerBound, upperBound, cut):
    k = len(graph.context.subalignments)
    gains = {}
    cutCost = 0
    
    for j in range(k):
        for node in range(lowerBound[j], upperBound[j]):
            gain = 0
            for i in range(k):
                for nbr, value in graph.nodeEdges[node][i]:
                    if nbr < lowerBound[i]:
                        continue
                    if nbr >= upperBound[i]:
                        break
                    
                    if (nbr < cut[i] and node < cut[j]) or (nbr >= cut[i] and node >= cut[j]):
                        gain = gain - value
                    else:
                        gain = gain + value
                        cutCost = cutCost + value
            gains[node] = gain 
    return gains, int(cutCost/2)               

def findNewBounds(graph, lowerBound, upperBound, cut, widthSumLimit):
    k = len(graph.context.subalignments)
    lowerSize = sum([cut[i] - lowerBound[i] for i in range(k)])
    upperSize = sum([upperBound[i] - cut[i] for i in range(k)])
    portionWidth = getPortionWidth(lowerBound, upperBound)
    
    limit = min(k, lowerSize + upperSize - 1)
    
    lowerMargin = int((lowerSize - upperSize + limit) * 0.5)
    upperMargin = int((upperSize - lowerSize + limit) * 0.5)
    
    #newLowerBound = [max(cut[i]-lowerMargin, lowerBound[i]) for i in range(k)]
    #newUpperBound = [min(cut[i]+upperMargin, upperBound[i]) for i in range(k)]
    newLowerBound = [max(cut[i]-lowerMargin, lowerBound[i], upperBound[i] - portionWidth + 1) for i in range(k)]
    newUpperBound = [min(cut[i]+upperMargin, upperBound[i], lowerBound[i] + portionWidth - 1) for i in range(k)]
    
    if widthSumLimit is not None:
        l1, l2, u1, u2 = None, None, None, None
        for i in range(k):
            l, u = cut[i]-lowerBound[i], upperBound[i]-cut[i]
            if l1 is None or l > cut[l1]-lowerBound[l1]:
                l1, l2 = i, l1
            elif l2 is None or l > cut[l2]-lowerBound[l2]:
                l2 = i
            if u1 is None or u > upperBound[u1]-cut[u1]:
                u1, u2 = i, u1
            elif u2 is None or u > upperBound[u2]-cut[u2]:
                u2 = i
        
        for i in range(k):
            if i == l1:
                newLowerBound[i] = max(newLowerBound[i], cut[l2] - lowerBound[l2] + upperBound[i] - widthSumLimit)
            else:
                newLowerBound[i] = max(newLowerBound[i], cut[l1] - lowerBound[l1] + upperBound[i] - widthSumLimit)
            if i == u1:
                newUpperBound[i] = min(newUpperBound[i], widthSumLimit + lowerBound[i] + cut[u2] - upperBound[u2])
            else:
                newUpperBound[i] = min(newUpperBound[i], widthSumLimit + lowerBound[i] + cut[u1] - upperBound[u1])
    
    return newLowerBound, newUpperBound

def updateGains(graph, lowerBound, upperBound, cut, gains, movedNodes, lowerUpdateBound, upperUpdateBound):
    k = len(graph.context.subalignments)
        
    for node in movedNodes:
        gains[node] = -1 * gains[node]
        asub, apos = graph.matSubPosMap[node]
        lowerUpdateBound[asub] = cut[asub]
        upperUpdateBound[asub] = cut[asub]
        
        for i in range(k):
            for nbr, value in graph.nodeEdges[node][i]:
                if nbr < lowerBound[i]:
                    continue
                if nbr >= upperBound[i]:
                    break
                
                if (nbr < cut[i] and node < cut[asub]) or (nbr >= cut[i] and node >= cut[asub]):
                    gains[nbr] = gains[nbr] - 2 * value
                else:
                    gains[nbr] = gains[nbr] + 2 * value
                
                if nbr < cut[i]:
                    lowerUpdateBound[i] = max(lowerUpdateBound[i], nbr+1)
                else:
                    upperUpdateBound[i] = min(upperUpdateBound[i], nbr)

def getHeapGainUpdateList(graph, newLowerBound, newUpperBound, lowerUpdateBound, upperUpdateBound):
    k = len(graph.context.subalignments)
    updateList = [[] for i in range(k)]      
    
    for i in range(k):
        for j in reversed(range(newLowerBound[i], lowerUpdateBound[i])):
            updateList[i].append(j)
        
        for j in range(upperUpdateBound[i], newUpperBound[i]):  
            updateList[i].append(j)
    
    return updateList

def updateHeapGainList(graph, cut, updateList, gains, heapGains, heapGainsVersions, heap, locked):
    k = len(graph.context.subalignments)      
    
    for i in range(k):
        for j in updateList[i]:
            heapGainsVersions[j] = heapGainsVersions.get(j, 0) + 1
            
            if j > cut[i]:
                heapGains[j] = heapGains[j-1] + gains[j]
            elif j == cut[i] or j == cut[i]-1:
                heapGains[j] = gains[j]
            else:
                heapGains[j] = heapGains[j+1] + gains[j]
            
            if j not in locked:
                heapq.heappush(heap, (-heapGains[j], j, heapGainsVersions[j]))

def getPortionWidth(lowerBound, upperBound):
    return max([u-l for u,l in zip(upperBound, lowerBound)])

def computeCutCost(graph, lowerBound, upperBound, cut):
    k = len(graph.context.subalignments)
    cutCost = 0
    
    for j in range(k):
        for node in range(lowerBound[j], upperBound[j]):
            for i in range(j, k):
                for nbr, value in graph.nodeEdges[node][i]:
                    if nbr < lowerBound[i]:
                        continue
                    if nbr >= upperBound[i]:
                        break
                    
                    if (nbr >= cut[i] and node < cut[j]) or (nbr < cut[i] and node >= cut[j]):
                        cutCost = cutCost + value
    return cutCost 


import os, sys, time, itertools

from witch_msa.configs import *
from witch_msa.gcmm.aligner import alignSubQueries, alignSubQueriesNew
from witch_msa.gcmm.loader import readCheckpointAlignments
from witch_msa.gcmm.task import *
from witch_msa.gcmm.callback import callback_queryAlignment

from witch_msa.helpers.alignment_tools import ExtendedAlignment

import concurrent.futures
from functools import partial
from tqdm import tqdm

'''
Function to yield the list of arguments for a job/task to perform.
'''
#def getTasks(remaining_indexes,
#        subset_query_names, subset_query_seqs, subset_weights):
#    for i in range(len(remaining_indexes)):
#        _i = remaining_indexes[i]
#        yield subset_query_names[_i], subset_query_seqs[_i], \
#                subset_weights[_i], _i

'''
Function to handle a single future returned from the main query alignment job
    future.result() should have two fields:
        0 - query alignment in ExtendedAlignment
        1 - query index
    Return: runtime in handling the future object
'''
#def handleFuture(future, success, ignored, retry, checkpoint_path,
#        i_retry=0):
#    s1 = time.time()
#    _query, _index = future.result()
#    if not _query and i_retry > 0:
#        retry.append(_index)
#    else:
#        # failed job indicated in the <witch-ng> or <default> pipelines,
#        # should be ignored in the output
#        if (not _query) or len(_query) == 0:
#            ignored.append(subset_query_names[_index])
#        else:
#            # write to checkpoint file
#            writeOneCheckpointAlignment(checkpoint_path, _query)
#            success.append(_query)
#    return time.time() - s1

'''
Function to handle returned future objects from query-alignments
(alignSubQueries* functions).
    b_retry - whether to retry some failed indexes (indicated by their return types)
    i_retry - the remaining number of retries allowed

    return: (1) a list of indexes for retry
            (2) the remaining number of retries
'''
#def handleFutures(futures, subset_query_names, success, ignored, i_retry,
#        checkpoint_path):
#    runtime_handle_result = 0.
#    retry = []
#    for future in tqdm(
#            concurrent.futures.as_completed(futures),
#            total=len(futures), **tqdm_styles):
#        s1 = time.time()
#        _query, _index = future.result()
#
#        # In the case of the <default> pipeline, there is a chance
#        # that the subprocess failed (e.g., due to timeout).
#        # Need to re-queue these jobs, possibly with a longer timeout threshold
#        if not _query and i_retry > 0:
#            retry.append(_index)
#        else:
#            # failed job indicated in the <witch-ng> or <default> pipelines,
#            # should be ignored in the output
#            if (not _query) or len(_query) == 0:
#                ignored.append(subset_query_names[_index])
#            else:
#                # write to checkpoint file
#                writeOneCheckpointAlignment(checkpoint_path, _query)
#                success.append(_query)
#        runtime_handle_result += time.time() - s1
#
#    return retry, max(0, i_retry - 1), runtime_handle_result


'''
Function to handle submitting and collecting results for all query alignments.
Also handles writing to checkpoints (so their runtimes are recorded)
'''
def submitAndCollectFutures(pool, lock, sid_to_query_names, sid_to_query_seqs,
        num_seq, index_to_hmm, taxon_to_weights,
        tmp_backbone_path, backbone_length):
    handle_runtime = 0.

    # allow for 1 retry if a query alignment failed (most likely due to 
    # running old-witch). Re-run with WITCH-ng pipeline
    i_retry = 1

    # items to return 
    queries, ignored = [], []

    subset_query_names = [sid_to_query_names[_i] for _i in range(num_seq)]
    subset_query_seqs = [sid_to_query_seqs[_i] for _i in range(num_seq)]
    remaining_indexes = [_i for _i in range(num_seq)]

    # set up checkpoints for query-alignments, and load in any finished
    # query alignments from a previous run
    checkpoint_path = Configs.outdir + '/checkpoint_alignments.txt.gz'

    # load from existing checkpoint path, if any exists
    checkpoint_queries = {}
    if os.path.exists(checkpoint_path) and os.stat(checkpoint_path).st_size > 0:
        print('\nFound existing checkpoint query alignments: {}'.format(
            checkpoint_path))
        checkpoint_queries = readCheckpointAlignments(checkpoint_path, pool, lock)
        print('\tSuccessfully read in {} checkpoint query alignments...'.format(
            len(checkpoint_queries)))

        # append checkpoint queries to queries
        queries = list(checkpoint_queries.values())
    
    # deal with queries that needs to be skipped (has no weight) or have
    # completed (from checkpoint)
    subset_weights, completed_indexes = [], []
    for _i in range(num_seq):
        _q = sid_to_query_names[_i]
        if _q in checkpoint_queries:
            completed_indexes.append(_i)
            subset_weights.append(tuple())
            continue

        if _q in taxon_to_weights:
            subset_weights.append(taxon_to_weights[_q])
        else:
            Configs.warning('{} does not have any weight/bitscore '.format(
                _q) + 'and will be ignored the final alignment.')
            subset_weights.append(tuple())
            ignored.append(_i)
    remaining_indexes = list(set(remaining_indexes).difference(set(completed_indexes)))
    remaining_indexes = list(set(remaining_indexes).difference(set(ignored)))

    # early return if no remaining indexes
    if len(remaining_indexes) == 0:
        return queries, ignored

    # Set up either WITCH or WITCH-ng's way of aligning the query sequence
    func_map = {'old-witch': alignSubQueries, 'witch-ng': alignSubQueriesNew}
    if func_map[Configs.mode]:
        func = partial(func_map[Configs.mode], tmp_backbone_path,
                backbone_length, index_to_hmm, lock, Configs.timeout)
    else:
        raise NotImplementedError

    mytasks = getTasksWithIndexes(remaining_indexes, subset_query_names,
            subset_query_seqs, subset_weights)
    success, _ignored, retry, handle_runtime = runTasks(
            func, pool, mytasks, len(remaining_indexes),
            i_retry=i_retry,
            max_concurrent_jobs=Configs.max_concurrent_jobs,
            callback_func=callback_queryAlignment,
            callback_args=[checkpoint_path]
            )
    queries.extend(success); ignored.extend(_ignored)

    # queue up re-try if any
    retry_func = partial(func_map['witch-ng'], tmp_backbone_path,
            backbone_length, index_to_hmm, lock, Configs.timeout)
    while len(retry) > 0 and i_retry > 0:
        i_retry -= 1
        retry_tasks = getTasksWithIndexes(retry, subset_query_names,
                subset_query_seqs, subset_weights)
        success, _ignored, retry, handle_runtime = runTasks(
                retry_func, pool, retry_tasks, len(retry),
                i_retry=i_retry,
                max_concurrent_jobs=Configs.max_concurrent_jobs,
                callback_func=callback_queryAlignment,
                callback_args=[checkpoint_path]
                )
        queries.extend(success); ignored.extend(_ignored)

    
    # try submitting jobs one-by-one and collect results
    #futures, success, retry = [], [], []
    # UPDATE 8.4.2023 - try limiting the total number of existing jobs
    #                 - by pausing submission after queue is full
    #with tqdm(total=len(remaining_indexes), **tqdm_styles) as pbar:
    #    tasks_to_do = getTasks(remaining_indexes, subset_query_names,
    #            subset_query_seqs, subset_weights)
    #    futures = {
    #        pool.submit(func, *task): task[-1]
    #        for task in itertools.islice(tasks_to_do, Configs.max_concurrent_jobs)
    #    }
    #    while futures:
    #        # wait for the next future to complete
    #        done, _ = concurrent.futures.wait(
    #                futures, return_when=concurrent.futures.FIRST_COMPLETED)

    #        # process results
    #        for future in done:
    #            # allow re-adding some failed jobs back to queue
    #            handle_runtime += handleFuture(future, success, ignored,
    #                    retry, checkpoint_path, i_retry=1)
    #            original_task = futures.pop(future)
    #            #print(f"Output of {original_task} is {future.result()}")
    #        pbar.update(len(done))

    #        # schedule the next set of tasks, no more than the number
    #        # of done to only have Configs.max_concurrent_jobs in the pool
    #        for task in itertools.islice(tasks_to_do, len(done)):
    #            future = pool.submit(func, *task)
    #            futures[future] = task[-1]
    
    # run retries if any exist
    # retry will use WTICH-ng's way (which presumably should be faster in the
    # case the subprocess reaches <timeout> seconds
    #while len(retry) > 0:
    #    print('\tQueuing up {} failed alignments, {} retries remaining...'.format(
    #        len(retry), i_retry))
    #    retry_futures = []
    #    retry_func = partial(func_map['witch-ng'], tmp_backbone_path,
    #            backbone_length, index_to_hmm, lock, Configs.timeout)
    #    for i in retry:
    #        retry_futures.append(pool.submit(retry_func, subset_query_names[i],
    #            subset_query_seqs[i], subset_weights[i], i))

    #    retry, i_retry, _runtime = handleFutures(retry_futures,
    #            subset_query_names, success, ignored, i_retry,
    #            checkpoint_path)
    #    handle_runtime += _runtime
    #queries.extend(success)

    Configs.runtime(' '.join(['(submitAndCollectFutures) Time to collect',
                'futures and write checkpoints (s):',
                str(handle_runtime)]))
    return queries, ignored



'''
Created on May 28, 2020

@author: Vlad
'''

import os
import random
import time

from align.decompose import initial_tree, kmh
from helpers import treeutils, sequenceutils
from configuration import Configs

'''
Handles the different ways to decompose the dataset into subsets.
The main way is to estimate a guide tree, then use PASTA's centroid edge decomposition
on the guide tree. Can also decompose randomly (for high speed on huge datasets).
'''

def decomposeSequences(context):
    time1 = time.time()
    
    if len(context.subsetPaths) > 0:
        Configs.log("Subset paths already provided, skipping decomposition..")
    
    elif len(context.subalignmentPaths) > 0:
        context.subsetPaths = context.subalignmentPaths
        Configs.log("Subalignment paths already provided, skipping decomposition..")
    
    else:
        subsetsDir = os.path.join(context.workingDir, "decomposition")
        context.subsetPaths = []
        n = 1
        while True:
            filePath = os.path.join(subsetsDir, "subset_{}.txt".format(n))
            if not os.path.exists(filePath):
                break
            Configs.log("Detected existing subset file {}".format(filePath))
            context.subsetPaths.append(filePath)
            n = n + 1
        
        if len(context.subsetPaths) == 0:
            buildDecomposition(context, subsetsDir)
    
    time2 = time.time()  
    Configs.log("Decomposed {} into {} subsets in {} sec..".format(context.sequencesPath, len(context.subsetPaths), time2-time1))

def buildDecomposition(context, subsetsDir):  
    if not os.path.exists(subsetsDir):
        os.makedirs(subsetsDir)  
    if context.unalignedSequences is None:
        context.unalignedSequences = sequenceutils.readFromFasta(context.sequencesPath, removeDashes=True)    
    
    if (Configs.decompositionStrategy == "random" or context.guideTree == "random") and Configs.outputPath == context.outputFile:
        context.subsetPaths = randomDecomposition(subsetsDir, context.unalignedSequences, Configs.decompositionMaxNumSubsets)
        
    elif Configs.decompositionStrategy == "kmh":
        Configs.log("Decomposing {} with KMH..".format(context.sequencesPath))
        Configs.log("Targetting {} subsets..".format(Configs.decompositionMaxNumSubsets))
        context.subsetPaths = kmh.buildSubsetsKMH(context, subsetsDir)
    
    else:
        guideTreePath  = initial_tree.buildInitialTree(context, subsetsDir, context.guideTree)
        Configs.log("Using target subset size of {}, and maximum number of subsets {}..".format(Configs.decompositionMaxSubsetSize, Configs.decompositionMaxNumSubsets))
        context.subsetPaths = treeutils.decomposeGuideTree(subsetsDir, context.sequencesPath, guideTreePath, 
                                                   Configs.decompositionMaxSubsetSize, Configs.decompositionMaxNumSubsets)        

def chooseSkeletonTaxa(sequences, skeletonSize, mode = "fulllength"):
    allTaxa = list(sequences.keys())
    
    if mode == "fulllength":
        seqLengths = [len(sequences[t].seq) for t in sequences]
        
        #topQuartile = numpy.quantile(seqLengths, 0.75)
        seqLengths.sort()
        topQuartile = seqLengths[int(0.75*(len(seqLengths)-1))]
        
        fullLength = []
        notFullLength = []
        for t in allTaxa:
            if abs(len(sequences[t].seq) - topQuartile) < 0.25 * topQuartile:
                fullLength.append(t)
            else:
                notFullLength.append(t) 
        
        random.shuffle(fullLength)
        random.shuffle(notFullLength)     
        allTaxa = fullLength + notFullLength
    else:
        random.shuffle(allTaxa)
        
    skeletonTaxa = allTaxa[:skeletonSize]
    remainingTaxa = allTaxa[skeletonSize:]
    return skeletonTaxa, remainingTaxa

def randomDecomposition(subsetsDir, sequences, numSubsets):
    allTaxa = list(sequences.keys())
    random.shuffle(allTaxa)
    
    taxonSubsets = [allTaxa[i :: numSubsets] for i in range(numSubsets)]
    subsetPaths = []
    for n, subset in enumerate(taxonSubsets):
        subsetPath = os.path.join(subsetsDir, "subset_{}.txt".format(n+1))
        subsetPaths.append(subsetPath)                    
        sequenceutils.writeFasta(sequences, subsetPath, subset) 
    return subsetPaths



from witch_msa.configs import Configs
from concurrent.futures import ProcessPoolExecutor

'''
Customized ProcessPoolExecutor class to handle callbacks and monitor current
progress in query alignments
'''
class WITCHProcessPoolExecutor(ProcessPoolExecutor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._running_jobs = 0
        self._submitted_jobs = 0
        self._finished_jobs = 0
    
    def submit(self, *args, **kwargs):
        future = super().submit(*args, **kwargs)
        self._running_jobs += 1
        self._submitted_jobs += 1
        #future.add_done_callback(self._worker_is_done)
        return future

    def _worker_is_done(self, future):
        self._running_jobs -= 1
        self._finished_jobs += 1
        print('Finished jobs: {}/{}'.format(
            self._finished_jobs, self._submitted_jobs), end='\r', flush=True)
    
    def get_pool_usage(self):
        return self._running_jobs
    
    def get_finished_jobs(self):
        return self._finished_jobs

def notifyError(location):
    print('Encountered an error at {}, check {}'.format(
        location, Configs.error_path))
    exit(1)
